from __future__ import print_function, division, absolute_import

import os
import shutil
import inspect

from functools import wraps
from six import string_types
from six.moves import builtins
from decorator import FunctionMaker
from collections import defaultdict, OrderedDict

import numpy as np

# to set the cache dir, set the environment CACHE_DIR
__cache_dir = os.environ.get("CACHE_DIR", os.path.join(os.path.expanduser('~'), '.odin_cache'))
_memory = None
# check cache_dir
if not os.path.exists(__cache_dir):
  os.mkdir(__cache_dir)
elif os.path.isfile(__cache_dir):
  raise ValueError("Invalid cache directory at path:" + __cache_dir)

# Don't use memmap anymore, carefull when cache big numpy ndarray results

def get_cache_memory():
  try:
    from joblib import Memory
  except ImportError as e:
    raise ImportError("Require 'joblib' to run cache utilities")
  global _memory
  if _memory is None:
    _memory = Memory(cachedir=__cache_dir,
                     mmap_mode=None, compress=False, verbose=0)
  return _memory


def get_cache_path():
  return __cache_dir


def clear_disk_cache():
  if os.path.exists(__cache_dir):
    shutil.rmtree(__cache_dir)


def cache_disk(function):
  """ Lazy evaluation of function, by storing the results to the disk,
  and not rerunning the function twice for the same arguments.

  It works by explicitly saving the output to a file and it is
  designed to work with non-hashable and potentially large input
  and output data types such as numpy arrays

  Note
  ----
  Any change in the function code will result re-cache

  Example
  -------
  >>> import numpy as np
  >>> @cache_memory
  >>> def doit(x):
  ...     y = np.random.rand(1208, 1208)
  ...     print("Function called:", np.sum(y))
  ...     return y
  >>> for _ in range(12):
  ...     print(np.sum(doit(8)))
  """
  def decorator_apply(dec, func):
    """Decorate a function by preserving the signature even if dec
    is not a signature-preserving decorator.
    This recipe is derived from
    http://micheles.googlecode.com/hg/decorator/documentation.html#id14
    """
    return FunctionMaker.create(
        func, 'return decorated(%(signature)s)',
        dict(decorated=dec(func)), __wrapped__=func)
  return decorator_apply(get_cache_memory().cache, function)


# ===========================================================================
# Cache
# ===========================================================================
__CACHE = defaultdict(lambda: ([], [])) #id(function) -> (KEY_ARGS, RET_VALUE)
__NO_ARGUMENT = '___NO_ARGUMENT___'


def __compare_cached_key(key, keylist):
  """Return index of match key if any key in the keylist match
  the given key"""
  try:
    return keylist.index(key)
  except ValueError:
    return None


def clear_mem_cache():
  __CACHE.clear()


def cache_memory(func, *attrs):
  '''Decorator. Caches the returned value and called arguments of
  a function.

  All the input and output are cached in the memory (i.e. RAM), and it
  requires hash-able inputs to compare the footprint of function.

  Parameters
  ----------
  attrs : str or list(str)
      list of object attributes in comparation for selecting cache value

  Note
  ----
  enable strict mode by specify "__strict__" in the `attrs`, this mode
  turns off caching by default but activated when "__cache__" appeared in
  the argument

  Example
  -------
  >>> class ClassName(object):
  >>>     def __init__(self, arg):
  >>>         super(ClassName, self).__init__()
  >>>         self.arg = arg
  >>>     @cache_memory('arg')
  >>>     def abcd(self, a):
  >>>         return np.random.rand(*a)
  >>>     def e(self):
  >>>         pass
  >>> x = c.abcd((10000, 10000))
  >>> x = c.abcd((10000, 10000)) # return cached value
  >>> c.arg = 'test'
  >>> x = c.abcd((10000, 10000)) # return new value
  '''
  strict_mode = False
  if not inspect.ismethod(func) and not inspect.isfunction(func):
    attrs = (func,) + attrs
    func = None
  # check if strict mode is enable
  if '__strict__' in attrs:
    strict_mode = True
    attrs = [a for a in attrs if a != '__strict__']
  # check if all attrs is string
  if builtins.any(not isinstance(i, string_types) for i in attrs):
    raise ValueError('Tracking attribute must be string represented name of'
                     ' attribute, but given attributes have types: {}'
                     ''.format(tuple(map(type, attrs))))
  # sort the attrs name so they always in unique order
  if len(attrs) > 0:
    attrs = sorted(attrs)

  def wrap_function(func):
    # ====== fetch arguments in order ====== #
    sign = inspect.signature(func)
    args_name = []
    args_defaults = OrderedDict()
    for n, p in sign.parameters.items():
      if p.kind in (inspect.Parameter.VAR_POSITIONAL,
                    inspect.Parameter.VAR_KEYWORD):
        continue
      args_name.append(n)
      if p.default != inspect.Parameter.empty:
        args_defaults[n] = p.default

    # ====== wraps the function ====== #
    @wraps(func)
    def wrapper(*args, **kwargs):
      # ====== check if strict_mode and caching enable ====== #
      if strict_mode:
        # __cache__ specified in args
        if any(isinstance(a, string_types) and a == '__cache__'
               for a in args):
          args = tuple([a for a in args
                        if not isinstance(a, string_types) or a != '__cache__'])
        # __cache__ specified in kwargs
        elif '__cache__' in kwargs:
          kwargs.pop('__cache__')
        # no cache just call the function
        else:
          return func(*args, **kwargs)
      # ====== additional arguments ====== #
      input_args = [__NO_ARGUMENT] * len(args_name)
      input_args[:len(args)] = args
      # merge default arguments
      for i, name in enumerate(args_name[len(args):]):
        if name in kwargs:
          input_args[len(args) + i] = kwargs[name]
        elif name in args_defaults:
          input_args[len(args) + i] = args_defaults[name]
        else:
          raise ValueError("Cannot find specified argument for "
              "argument with name: %s" % name)
      # ====== create cache_key ====== #
      # custom attribute
      object_attrs = [getattr(args[0], k) for k in attrs
                      if hasattr(args[0], k)]
      cache_key = input_args + object_attrs
      cache_key = [id(k) if isinstance(k, np.ndarray) else k
                   for k in cache_key]
      # ====== check cache ====== #
      key_list = __CACHE[id(func)][0]
      value_list = __CACHE[id(func)][1]
      match_index = __compare_cached_key(cache_key, key_list)
      # get old cached value
      if match_index is not None:
        return value_list[match_index]
      # call the function to get new cached value
      else:
        value = func(*args, **kwargs)
        key_list.append(cache_key)
        value_list.append(value)
        return value
    return wrapper

  # return wrapped function
  if func is None:
    return wrap_function
  return wrap_function(func)



# ===========================================================================
# Class handle extreme large numpy ndarray
# ===========================================================================
from __future__ import print_function, division, absolute_import

import os
import re
import marshal
from math import ceil
from six import add_metaclass
from contextlib import contextmanager
from abc import ABCMeta, abstractmethod, abstractproperty
from six.moves import range, zip, zip_longest
from collections import OrderedDict, defaultdict

import numpy as np
from numpy import ndarray

from odin.utils.decorators import autoattr
from odin.utils import (struct, as_tuple, cache_memory, is_string,
                        is_number, is_primitives, axis_normalize)

__all__ = [
  'as_data',
  'open_hdf5',
  'close_all_hdf5',
  'get_all_hdf_dataset',

  'Data',
  'DataConcat',
  'DataGroup',
  'DataCopy',
  'NdarrayData',
  'MmapData',
  'Hdf5Data',
  'DataIterator',
]

# maximum batch size in bytes when performing apply
MAX_BUFFER_SIZE = 100 * 1024 * 1024 # 100 MB


# ===========================================================================
# Helper function
# ===========================================================================
def as_data(x, copy=False):
  """ make sure x is instance Data
  Parameters
  ----------
  copy : bool
    if True, copy the Data object to a separated instance
    to prevent further unexpected change to the configuration
  """
  if isinstance(x, Data):
    if copy:
      if 'odin.fuel.feeder.Feeder' in str(type(x)):
        raise ValueError("Cannot copy `Feeder` data type.")
      return DataCopy(x)
    else:
      return x
  if isinstance(x, np.ndarray):
    return NdarrayData(x)
  if isinstance(x, (tuple, list)):
    if is_primitives(x[0], inc_ndarray=False): # given value for array
      return NdarrayData(np.asarray(x))
    return DataGroup(data=x)
  raise ValueError('Cannot create Data object from given object:{}'.format(x))


def _get_chunk_size(shape, size):
  if isinstance(size, int):
    return (2**int(np.ceil(np.log2(size))),) + shape[1:]
  elif size is None:
    return False
  return True


def _validate_operate_axis(axis):
  ''' as we iterate over first dimension, it is prerequisite to
  have 0 in the axis of operator
  '''
  if not isinstance(axis, (tuple, list)):
    axis = [axis]
  axis = tuple(int(i) for i in axis)
  if 0 not in axis:
    raise ValueError('Expect 0 in the operating axis because we always'
                     ' iterate data over the first dimension.')
  return axis


# x can be percantage or number of samples
_apply_approx = lambda n, x: int(round(n * x)) if x < 1. + 1e-12 else int(x)


# ===========================================================================
# Data
# ===========================================================================
@add_metaclass(ABCMeta)
class Data(object):

  """ Note for overriding `Data` class:
  `_data` will always be a tuple.
  """

  def __init__(self, data, read_only):
    # batch information
    self._batch_size = 256
    self._start = 0.
    self._end = 1.
    self._seed = None
    self._shuffle_level = 0
    # ====== main data ====== #
    # object that have shape, dtype ...
    self._data = as_tuple(data)
    if isinstance(data, (tuple, list)):
      self._is_data_list = True
    else:
      self._is_data_list = False
    self._read_only = bool(read_only)
    # ====== special flags ====== #
    # to detect if cPickle called with protocol >= 2
    self._new_args_called = False
    # flag show that array valued changed
    self._status = 0

  def resize(self, new_length):
    raise NotImplementedError

  # ==================== abstract ==================== #
  @abstractmethod
  def _restore_data(self, info):
    raise NotImplementedError

  @abstractproperty
  def data_info(self):
    """This property return info for restore data object after
    pickling.

    NOTE
    ----
    The return values from this function will be introduce as
    input argument to `_restore_data`
    """
    raise NotImplementedError

  # ==================== basic ==================== #
  def flush(self):
    pass

  def close(self):
    pass

  # ==================== basic properties ==================== #
  @property
  def is_data_list(self):
    """ Return whether this Data was created as a list of given Data
    or just a single Data.
    This will affects the return result of `shape` (i.e.
    whether it should be a list of shape tuple, or just single shape)
    """
    return self._is_data_list

  @property
  def read_only(self):
    return self._read_only

  @property
  def batch_size(self):
    return self._batch_size

  @property
  def data(self):
    return self._data

  @property
  def array(self):
    """Convert any data manipulated by this object to
    numpy.ndarray"""
    a = tuple([dat[:] for dat in self._data])
    return a if self.is_data_list else a[0]

  def tolist(self):
    a = tuple([dat.tolist() for dat in self._data])
    return a if self.is_data_list else a[0]

  # ==================== For pickling ==================== #
  @property
  def new_args(self):
    """ This positional arguments will be presented during
    unpickling. """
    return ()

  def __getstate__(self):
    if not self._new_args_called:
      raise RuntimeError(
          "You must use argument `protocol=cPickle.HIGHEST_PROTOCOL` "
          "when using `pickle` or `cPickle` to be able pickling Data.")
    self._new_args_called = False
    return (self._batch_size, self._start, self._end,
            self._seed, self._shuffle_level,
            self._is_data_list, self.data_info,
            self._read_only)

  def __setstate__(self, states):
    (self._batch_size, self._start, self._end,
        self._seed, self._shuffle_level,
        self._is_data_list, data_info,
        self._read_only) = states
    self._status = 0
    self._new_args_called = False
    # ====== restore data ====== #
    self._restore_data(data_info)
    if not hasattr(self, '_data') or self._data is None:
      raise RuntimeError("The `_data` attribute is None, and have not "
                         "been restored after pickling, you must properly "
                         "implement function `restore_data` for class '%s'"
                         % self.__class__.__name__)

  def __getnewargs__(self):
    self._new_args_called = True
    return as_tuple(self.new_args)

  # ==================== properties ==================== #
  @property
  def ndim(self):
    if self.is_data_list:
      return tuple([len(i) for i in self.shape])
    return len(self.shape)

  @property
  def shape(self):
    s = tuple([i.shape for i in self._data])
    return s if self.is_data_list else s[0]

  @property
  def iter_len(self):
    length = self.__len__()
    start = _apply_approx(length, self._start)
    end = _apply_approx(length, self._end)
    return end - start

  def __len__(self):
    """ len always return 1 number """
    shape = self.shape
    if is_number(shape[0]):
      return shape[0]
    return self.shape[0][0]

  @property
  def T(self):
    x = tuple([i.T for i in self._data])
    return x if self.is_data_list else x[0]

  @property
  def dtype(self):
    x = tuple([np.dtype(i.dtype) for i in self._data])
    return x if self.is_data_list else x[0]

  @contextmanager
  def set_batch_context(self, batch_size=None, seed=-1, start=None, end=None,
                        shuffle_level=None):
    _batch_size = self._batch_size
    _seed = self._seed
    _start = self._start
    _end = self._end
    _shuffle_level = self._shuffle_level
    # temporary batch configuration
    self.set_batch(batch_size=_batch_size if batch_size is None else batch_size,
        seed=_seed if seed == -1 else seed,
        start=_start if start is None else start,
        end=_end if end is None else end,
        shuffle_level=_shuffle_level if shuffle_level is None else shuffle_level)
    yield self
    # reset batch to original batch configuration
    self.set_batch(batch_size=_batch_size, seed=_seed,
                   start=_start, end=_end,
                   shuffle_level=_shuffle_level)

  def set_batch(self, batch_size=None, seed=-1, start=None, end=None,
                shuffle_level=None):
    """
    Parameters
    ----------
    batch_size: int
        size of each batch return when iterate this Data
    seed: None, int
        if None, no shuffling is performed while iterating,
        if < 0, do not change the current seed
        if >= 0, enable randomization with given seed
    start: int, float
        if int, start indicates the index of starting data points to
        iterate. If float, start is the percentage of data to start.
    end: int, float
        ending point of the interation
    shuffle_level: int
        0: only shuffle the order of each batch
        1: shuffle the order of batches and inside each batch as well.
        2: includes level 0 and 1, and custom shuffling (strongest form)
    """
    if isinstance(batch_size, int) and batch_size > 0:
      self._batch_size = batch_size
    if seed is None or seed >= 0:
      self._seed = seed
    if start is not None and start > 0. - 1e-12:
      self._start = start
    if end is not None and end > 0. - 1e-12:
      self._end = end
    if shuffle_level is not None:
      self._shuffle_level = min(max(int(shuffle_level), 0), 2)
    return self

  # ==================== Slicing methods ==================== #
  def __getitem__(self, key):
    x = tuple([dat.__getitem__(key) for dat in self.data])
    return x if self.is_data_list else x[0]

  @autoattr(_status=lambda x: x + 1)
  def __setitem__(self, x, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode.")
    # perform
    for dat in self._data:
      dat.__setitem__(x, y)
    return self

  # ==================== iteration ==================== #
  def __iter__(self):
    def create_iteration():
      batch_size = int(self._batch_size)
      length = self.__len__()
      seed = self._seed
      shuffle_level = int(self._shuffle_level)
      self._seed = None
      # custom batch_size
      start = _apply_approx(length, self._start)
      end = _apply_approx(length, self._end)
      if not (0 <= start < end <= length):
        raise ValueError('`start`={} or `end`={} excess `data_size`={}'
                         ''.format(start, end, length))
      # ====== create batch ====== #
      idx = list(range(start, end, batch_size))
      if idx[-1] < end:
        idx.append(end)
      idx = list(zip(idx, idx[1:]))
      # ====== shuffling the batch ====== #
      if seed is not None:
        rand = np.random.RandomState(seed=seed)
        rand.shuffle(idx)
      else:
        rand = None
      # this dummy return to make everything initialized
      yield None
      # ====== start the iteration ====== #
      for start, end in idx:
        # [i[start:end] for i in self._data]
        x = self.__getitem__(slice(start, end))
        # shuffle with higher level
        if shuffle_level > 0 and rand is not None:
          permu = rand.permutation(end - start)
          x = tuple([i[permu] for i in x]) \
              if isinstance(x, (tuple, list)) else x[permu]
        yield x
    # ====== create, init, and return the iteration ====== #
    it = create_iteration()
    next(it)
    return it

  # ==================== Strings ==================== #
  def __str__(self):
    s = "<%s: " % self.__class__.__name__
    for dat in self.data:
      s += "(%s dtype:%s shape:%s)" % \
          (dat.__class__.__name__, dat.dtype.name, dat.shape)
    s += '>'
    return s

  def __repr__(self):
    return self.__str__()

  # ==================== manipulation ==================== #
  @autoattr(_status=lambda x: x + 1)
  def append(self, *arrays):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    accepted_arrays = []
    add_size = 0
    # ====== check if shape[1:] matching ====== #
    for a, d in zip(arrays, self._data):
      if hasattr(a, 'shape'):
        if a.shape[1:] == d.shape[1:]:
          accepted_arrays.append(a)
          add_size += a.shape[0]
      else:
        accepted_arrays.append(None)
    # ====== resize ====== #
    old_size = self.__len__()
    # special case, Mmap is init with temporary size = 1 (all zeros),
    # NOTE: risky to calculate sum of big array here
    if old_size == 1 and \
    sum(np.sum(np.abs(d[:])) for d in self._data) == 0.:
      old_size = 0
    # resize and append data
    self.resize(old_size + add_size) # resize only once will be faster
    # ====== update values ====== #
    for a, d in zip(accepted_arrays, self._data):
      if a is not None:
        d[old_size:old_size + a.shape[0]] = a
    return self

  # ==================== high-level operators ==================== #
  @cache_memory('_status')
  def apply(self, f, f_merge=lambda x: x):
    """ Iteratively execute a list of function `f` on the Data.

    BigData instance store large dataset that need to be iterate
    over to perform any operators.

    Parameters
    ----------
    f: call-able
        a function applied on each returned batch
    f_merge: call-able
        a function applied on all batch results
    """
    # ====== validate arguments ====== #
    if not hasattr(f, '__call__'):
      raise ValueError("`f` must be call-able function.")
    if not hasattr(f_merge, '__call__'):
      raise ValueError("`f_merge` must be call-able function.")
    # ====== Processing ====== #
    results = []
    old_batch_size = self._batch_size
    old_seed = self._seed
    old_start = self._start
    old_end = self._end
    bs = MAX_BUFFER_SIZE / sum(np.prod(i.shape[1:]) * i.dtype.itemsize
                               for i in self._data)
    # less than million data points, not a big deal
    for X in self.set_batch(batch_size=bs, start=0., end=1., seed=None):
      results.append(f(X))
    self.set_batch(batch_size=old_batch_size, start=old_start, end=old_end,
                   seed=old_seed)
    # ====== reset and return ====== #
    return f_merge(results)

  @cache_memory('_status')
  def sum(self, axis=None):
    if axis is None or 0 in as_tuple(axis, t=int):
      y = self.apply(
          f=lambda x: [i.sum(axis=axis)
                       for i in as_tuple(x)],
          f_merge=lambda x: [sum(j[i] for j in x)
                             for i in range(len(x[0]))])
    else:
      y = [i.sum(axis=axis)
           for i in self._data]
    return y if self.is_data_list else y[0]

  @cache_memory('_status')
  def cumsum(self, axis=None):
    x = [i.cumsum(axis) for i in self._data]
    return x if self.is_data_list else x[0]

  @cache_memory('_status')
  def sum2(self, axis=None):
    if axis is None or 0 in as_tuple(axis, t=int):
      y = self.apply(
          f=lambda x: [i.__pow__(2).sum(axis=axis)
                       for i in as_tuple(x)],
          f_merge=lambda x: [sum(j[i] for j in x)
                             for i in range(len(x[0]))])
    else:
      y = [i.__pow__(2).sum(axis=axis)
           for i in self._data]
    return y if self.is_data_list else y[0]

  @cache_memory('_status')
  def pow(self, y):
    x = [i.__pow__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  @cache_memory('_status')
  def min(self, axis=None):
    if axis is None or 0 in as_tuple(axis, t=int):
      y = self.apply(
          f=lambda x: [i.min(axis=axis)
                       for i in as_tuple(x)],
          f_merge=lambda x: [np.min([j[i] for j in x], axis=0)
                             for i in range(len(x[0]))])
    else:
      y = [i.min(axis=axis)
           for i in self._data]
    return y if self.is_data_list else y[0]

  @cache_memory('_status')
  def argmin(self, axis=None):
    x = [i.argmin(axis) for i in self._data]
    return x if self.is_data_list else x[0]

  @cache_memory('_status')
  def max(self, axis=None):
    if axis is None or 0 in as_tuple(axis, t=int):
      y = self.apply(
          f=lambda x: [i.max(axis=axis)
                       for i in as_tuple(x)],
          f_merge=lambda x: [np.max([j[i] for j in x], axis=0)
                             for i in range(len(x[0]))])
    else:
      y = [i.max(axis=axis)
           for i in self._data]
    return y if self.is_data_list else y[0]

  @cache_memory('_status')
  def argmax(self, axis=None):
    x = [i.argmax(axis) for i in self._data]
    return x if self.is_data_list else x[0]

  # ==================== probs and stats ==================== #
  @cache_memory('_status')
  def mean(self, axis=None):
    sum1 = as_tuple(self.sum(axis))
    results = []
    for s1, dat in zip(sum1, self._data):
      shape = dat.shape
      if axis is None:
        n = np.prod(shape)
      else:
        n = np.prod([shape[i] for i in as_tuple(axis, t=int)])
      results.append(s1 / n)
    return results if self.is_data_list else results[0]

  @cache_memory('_status')
  def var(self, axis=None):
    sum1 = as_tuple(self.sum(axis))
    sum2 = as_tuple(self.sum2(axis))

    results = []
    for s1, s2, dat in zip(sum1, sum2, self._data):
      shape = dat.shape
      if axis is None:
        n = np.prod(shape)
      else:
        n = np.prod([shape[i] for i in as_tuple(axis, t=int)])
      results.append((s2 - np.power(s1, 2) / n) / n)
    return results if self.is_data_list else results[0]

  @cache_memory('_status')
  def std(self, axis=None):
    s = [np.sqrt(v) for v in as_tuple(self.var(axis))]
    return s if self.is_data_list else s[0]

  # ==================== low-level operator ==================== #
  def __add__(self, y):
    x = [i.__add__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  def __sub__(self, y):
    x = [i.__sub__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  def __mul__(self, y):
    x = [i.__mul__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  def __div__(self, y):
    x = [i.__div__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  def __floordiv__(self, y):
    x = [i.__floordiv__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  def __pow__(self, y):
    x = [i.__pow__(y) for i in self._data]
    return x if self.is_data_list else x[0]

  def __neg__(self):
    x = [i.__neg__() for i in self._data]
    return x if self.is_data_list else x[0]

  def __pos__(self):
    x = [i.__pos__() for i in self._data]
    return x if self.is_data_list else x[0]

  # ==================== update function ==================== #
  @autoattr(_status=lambda x: x + 1)
  def __iadd__(self, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    for i in self._data:
      i.__iadd__(y)

  @autoattr(_status=lambda x: x + 1)
  def __isub__(self, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    for i in self._data:
      i.__isub__(y)

  @autoattr(_status=lambda x: x + 1)
  def __imul__(self, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    for i in self._data:
      i.__imul__(y)

  @autoattr(_status=lambda x: x + 1)
  def __idiv__(self, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    for i in self._data:
      i.__idiv__(y)

  @autoattr(_status=lambda x: x + 1)
  def __ifloordiv__(self, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    for i in self._data:
      i.__ifloordiv__(y)

  @autoattr(_status=lambda x: x + 1)
  def __ipow__(self, y):
    if self.read_only:
      raise RuntimeError("This Data is set in read-only mode")
    for i in self._data:
      i.__ipow__(y)

# ===========================================================================
# Utils
# ===========================================================================
class DataConcat(Data):
  """ Concatenate two Data while iterating over them
  """

  def __init__(self, data, axis=-1):
    data = as_tuple(data)
    if len(data) < 2:
      raise ValueError("2 or more Data must be given to `DataConcat`")
    if axis == 0:
      raise ValueError("Cannot concatenate axis=0")
    if len(set(d.ndim for d in data)) > 2:
      raise ValueError("All Data must have the same number of dimension (i.e. `ndim`)")
    if len(set(d.shape[0] for d in data)) > 2:
      raise ValueError("All Data must have the same length (i.e. first dimension)")
    super(DataConcat, self).__init__(data, read_only=True)
    self._is_data_list = False
    self._axis = axis_normalize(int(axis), ndim=data[0].ndim)

  @property
  def data_info(self):
    return self._data, self._axis

  def _restore_data(self, info):
    self._data, self._axis = info

  @property
  def array(self):
    """Convert any data manipulated by this object to
    numpy.ndarray"""
    return np.concatenate([dat[:] for dat in self._data],
                          axis=self._axis)

  def tolist(self):
    return self.array.tolist()

  @property
  def shape(self):
    all_shapes = [d.shape for d in self.data]
    new_shape = [all_shapes[0][0]]
    ndim = len(all_shapes[0])
    for i in range(1, ndim):
      if i == self._axis:
        new_shape.append(sum(s[i] for s in all_shapes))
      else:
        new_shape.append(all_shapes[0][i])
    return tuple(new_shape)

  def __getitem__(self, key):
    return np.concatenate([d.__getitem__(key) for d in self.data],
                          axis=self._axis)

class DataGroup(Data):
  """ Group serveral Data together and return each Data
  separately during iteration

  This class is always considered as list of Data
  """

  def __init__(self, data):
    for d in as_tuple(data):
      if not isinstance(d, (np.ndarray, Data)):
        raise ValueError('`data` can only be instance of numpy.ndarray or '
                         'odin.fuel.data.Data, but given type: %s' % str(type(d)))
    data = [NdarrayData(d) if isinstance(d, np.ndarray) else d
            for d in as_tuple(data)]
    if len(set(len(d) for d in data)) > 1:
      raise ValueError("All data in given data list must have the same length, "
                       "but given: %s" % str([len(d) for d in data]))
    super(DataGroup, self).__init__(data, read_only=True)

  @property
  def data_info(self):
    return self._data

  def _restore_data(self, info):
    self._data = info

class DataCopy(Data):
  """ Simple copy that contain original version of Data
  """

  def __init__(self, data):
    if not isinstance(data, Data):
      raise ValueError("`data` must be instance of odin.fuel.data.Data but "
                       "given type: %s" % str(type(data)))
    # ====== special case DataGroup ====== #
    if isinstance(data, DataGroup):
      super(DataCopy, self).__init__(data._data, read_only=True)
    else:
      super(DataCopy, self).__init__(data, read_only=True)
    # ====== copy information ====== #
    self._is_data_list = data._is_data_list
    self._batch_size = data._batch_size
    self._start = data._start
    self._end = data._end
    self._seed = data._seed
    self._shuffle_level = data._shuffle_level

  @property
  def data_info(self):
    return self._data

  def _restore_data(self, info):
    self._data = info

class NdarrayData(Data):
  """ Simple wrapper for `numpy.ndarray` """

  def __init__(self, array):
    if not isinstance(array, np.ndarray):
      raise ValueError('array must be instance of numpy ndarray')
    super(NdarrayData, self).__init__(array, read_only=False)

  @property
  def data_info(self):
    return self._data

  def _restore_data(self, info):
    self._data = info

  # ==================== abstract ==================== #
  def resize(self, new_length):
    shape = self._data[0].shape
    new_shape = (new_length,) + shape[1:]
    return self._data[0].resize(new_shape)

# ===========================================================================
# Memmap Data object
# ===========================================================================
MAX_OPEN_MMAP = 120

def _aligned_memmap_offset(dtype):
  header_size = len(MmapData.HEADER) + 8 + MmapData.MAXIMUM_HEADER_SIZE
  type_size = np.dtype(dtype).itemsize
  n = np.ceil(header_size / type_size)
  return int(n * type_size)

class MmapData(Data):

  """Create a memory-map to an array stored in a *binary* file on disk.

  Memory-mapped files are used for accessing small segments of large files
  on disk, without reading the entire file into memory.  Numpy's
  memmap's are array-like objects.  This differs from Python's ``mmap``
  module, which uses file-like objects.

  Parameters
  ----------
  path : str
      The file name or file object to be used as the array data buffer.
  dtype : data-type, optional
      The data-type used to interpret the file contents.
      Default is `uint8`.
  shape : tuple, optional
      The desired shape of the array. If ``mode == 'r'`` and the number
      of remaining bytes after `offset` is not a multiple of the byte-size
      of `dtype`, you must specify `shape`. By default, the returned array
      will be 1-D with the number of elements determined by file size
      and data-type.

  Note
  ----
  This class always read MmapData with mode=r+
  """
  _INSTANCES = OrderedDict()
  HEADER = b'mmapdata'
  MAXIMUM_HEADER_SIZE = 486

  @staticmethod
  def read_header(path, read_only, return_file):
    """ return: dtype, shape
    Necessary information to create numpy.memmap
    """
    f = open(path, mode='rb' if read_only else 'rb+')
    # ====== check header signature ====== #
    try:
      if f.read(len(MmapData.HEADER)) != MmapData.HEADER:
        raise Exception
    except Exception as e:
      f.close()
      raise Exception('Invalid header for MmapData.')
    # ====== 8 bytes for size of info ====== #
    try:
      size = int(f.read(8))
      dtype, shape = marshal.loads(f.read(size))
    except Exception as e:
      f.close()
      raise Exception('Error reading memmap data file: %s' % str(e))
    # ====== return file object ====== #
    if return_file:
      return dtype, shape, f
    f.close()
    return dtype, shape

  def __new__(clazz, *args, **kwargs):
    path = kwargs.get('path', None)
    if path is None:
      path = args[0]
    if not is_string(path):
      raise ValueError("`path` for MmapData must be string, but given "
                       "object with type: %s" % type(path))
    path = os.path.abspath(path)
    # Found old instance
    if path in MmapData._INSTANCES:
      return MmapData._INSTANCES[path]
    # new MmapData
    # ====== increase memmap count ====== #
    if len(MmapData._INSTANCES) + 1 > MAX_OPEN_MMAP:
      raise ValueError('Only allowed to open maximum of {} memmap file'.format(MAX_OPEN_MMAP))
    # ====== create new instance ====== #
    new_instance = super(MmapData, clazz).__new__(clazz)
    MmapData._INSTANCES[path] = new_instance
    return new_instance

  def __init__(self, path, dtype='float32', shape=None,
               read_only=False):
    # validate path
    path = os.path.abspath(path)
    # ====== check shape info ====== #
    if shape is not None:
      if not isinstance(shape, (tuple, list, np.ndarray)):
        shape = (shape,)
      shape = tuple([0 if i is None or i < 0 else i for i in shape])
    # ====== read exist file ====== #
    if os.path.exists(path):
      dtype, shape, f = MmapData.read_header(path,
                                             read_only=read_only,
                                             return_file=True)
    # ====== create new file ====== #
    else:
      if dtype is None or shape is None:
        raise Exception("First created this MmapData, `dtype` and "
                        "`shape` must NOT be None.")
      f = open(path, 'wb+')
      f.write(MmapData.HEADER)
      dtype = str(np.dtype(dtype))
      if isinstance(shape, np.ndarray):
        shape = shape.tolist()
      if not isinstance(shape, (tuple, list)):
        shape = (shape,)
      _ = marshal.dumps([dtype, shape])
      size = len(_)
      if size > MmapData.MAXIMUM_HEADER_SIZE:
        raise Exception('The size of header excess maximum allowed size '
                        '(%d bytes).' % MmapData.MAXIMUM_HEADER_SIZE)
      size = '%8d' % size
      f.write(size.encode())
      f.write(_)
    # ====== assign attributes ====== #
    self._file = f
    self._path = path
    data = np.memmap(f, dtype=dtype, shape=shape,
                     mode = 'r' if read_only else 'r+',
                     offset=_aligned_memmap_offset(dtype))
    # finally call super initialize
    super(MmapData, self).__init__(data=data, read_only=read_only)

  @property
  def data_info(self):
    return self._path

  @property
  def path(self):
    return self._path

  @property
  def new_args(self):
    return (self._path,)

  def _restore_data(self, info):
    # info here is the path
    dtype, shape, f = MmapData.read_header(info, mode='r+',
                                           return_file=True)
    self._file = f
    self._data = np.memmap(f, dtype=dtype, shape=shape, mode='r+',
                           offset=_aligned_memmap_offset(dtype))
    self._path = info

  def flush(self):
    if self.read_only:
      return
    self._data[0].flush()

  def close(self):
    # Check if exist global instance
    if self.data_info in MmapData._INSTANCES:
      del MmapData._INSTANCES[self.data_info]
      # flush in read-write mode
      if not self.read_only:
        self.flush()
      # close mmap and file
      self._data[0]._mmap.close()
      del self._data
      self._file.close()

  # ==================== properties ==================== #
  def __str__(self):
    return '<MMAP "%s": %s "%s">' % \
    (self.data_info, self.shape, self.dtype)

  # ==================== Save ==================== #
  def resize(self, new_length):
    if self.read_only:
      return
    # ====== local files ====== #
    f = self._file
    mmap = self._data[0]
    old_length = mmap.shape[0]
    # ====== check new shape ====== #
    if new_length < old_length:
      raise ValueError('Only support extend memmap, and do not shrink the memory')
    # nothing to resize
    elif new_length == old_length:
      return self
    # ====== flush previous changes ====== #
    # resize by create new memmap and also rename old file
    shape = (new_length,) + mmap.shape[1:]
    dtype = mmap.dtype.name
    # rewrite the header
    f.seek(len(MmapData.HEADER))
    meta = marshal.dumps([dtype, shape])
    size = '%8d' % len(meta)
    f.write(size.encode(encoding='utf-8'))
    f.write(meta)
    f.flush()
    # extend the memmap
    mmap._mmap.close()
    del self._data
    mmap = np.memmap(self.data_info, dtype=dtype, shape=shape,
                     mode='r+',
                     offset=_aligned_memmap_offset(dtype))
    self._data = (mmap,)
    return self

# ===========================================================================
# Hdf5 Data object
# ===========================================================================
try:
  import h5py
except:
  pass


def get_all_hdf_dataset(hdf, fileter_func=None, path='/'):
  res = []
  # init queue
  q = queue()
  for i in hdf[path].keys():
    q.put(i)
  # get list of all file
  while not q.empty():
    p = q.pop()
    if 'Dataset' in str(type(hdf[p])):
      if fileter_func is not None and not fileter_func(p):
        continue
      res.append(p)
    elif 'Group' in str(type(hdf[p])):
      for i in hdf[p].keys():
        q.put(p + '/' + i)
  return res


_HDF5 = {}


def open_hdf5(path, read_only=False):
  '''
  Parameters
  ----------
  mode : one of the following options
      +------------------------------------------------------------+
      |r        | Readonly, file must exist                        |
      +------------------------------------------------------------+
      |r+       | Read/write, file must exist                      |
      +------------------------------------------------------------+
      |w        | Create file, truncate if exists                  |
      +------------------------------------------------------------+
      |w- or x  | Create file, fail if exists                      |
      +------------------------------------------------------------+
      |a        | Read/write if exists, create otherwise (default) |
      +------------------------------------------------------------+

  check : bool
      if enable, only return openned files, otherwise, None

  Note
  ----
  If given file already open in read mode, mode = 'w' will cause error
  (this is good error and you should avoid this situation)

  '''
  key = os.path.abspath(path)
  mode = 'r' if read_only else 'a'

  if key in _HDF5:
    f = _HDF5[key]
    if 'Closed' in str(f):
      f = h5py.File(path, mode=mode)
  else:
    f = h5py.File(path, mode=mode)
    _HDF5[key] = f
  return f


def close_all_hdf5():
  import gc
  for obj in gc.get_objects():   # Browse through ALL objects
    if isinstance(obj, h5py.File):   # Just HDF5 files
      try:
        obj.close()
      except:
        pass # Was already closed


class Hdf5Data(Data):

  def __init__(self, dataset, hdf=None, dtype=None, shape=None):
    super(Hdf5Data, self).__init__()
    raise Exception("Hdf5Data is under-maintanance!")
    # default chunks size is 32 (reduce complexity of the works)
    self._chunk_size = 32
    if isinstance(hdf, str):
      hdf = open_hdf5(hdf)
    if hdf is None and not isinstance(dataset, h5py.Dataset):
      raise ValueError('Cannot initialize dataset without hdf file')

    if isinstance(dataset, h5py.Dataset):
      self._data = dataset
      self._hdf = dataset.file
    else:
      if dataset not in hdf: # not created dataset
        if dtype is None or shape is None:
          raise ValueError('dtype and shape must be specified if '
                           'dataset has not created in hdf5 file.')
        shape = tuple([0 if i is None else i for i in shape])
        hdf.create_dataset(dataset, dtype=dtype,
            chunks=_get_chunk_size(shape, self._chunk_size),
            shape=shape, maxshape=(None, ) + shape[1:])

      self._data = hdf[dataset]
      if shape is not None and self._data[0].shape[1:] != shape[1:]:
        raise ValueError('Shape mismatch between predefined dataset '
                         'and given shape, {} != {}'
                         ''.format(shape, self._data[0].shape))
      self._hdf = hdf

  # ==================== properties ==================== #
  @property
  def path(self):
    return self._hdf.filename

  @property
  def name(self):
    _ = self._data[0].name
    if _[0] == '/':
      _ = _[1:]
    return _

  @property
  def hdf5(self):
    return self._hdf

  # ==================== Save ==================== #
  def resize(self, shape):
    if self._hdf.mode == 'r':
      return

    if not isinstance(shape, (tuple, list)):
      shape = (shape,)
    if any(i != j for i, j in zip(shape[1:], self._data[0].shape[1:])):
      raise ValueError('Resize only support the first dimension, but '
                       '{} != {}'.format(shape[1:], self._data[0].shape[1:]))
    if shape[0] < self._data[0].shape[0]:
      raise ValueError('Only support extend memmap, and do not shrink the memory')
    elif shape[0] == self._data[0].shape[0]:
      return self

    self._data[0].resize(shape[0], axis=0)
    return self

  def flush(self):
    try:
      if self._hdf.mode == 'r':
        return
      self._hdf.flush()
    except Exception:
      pass

  def close(self):
    try:
      self._hdf.close()
    except Exception:
      pass


# ===========================================================================
# data iterator
# ===========================================================================
def _approximate_continuos_by_discrete(distribution):
  '''original distribution: [ 0.47619048  0.38095238  0.14285714]
     best approximated: [ 5.  4.  2.]
  '''
  if len(distribution) == 1:
    return distribution

  inv_distribution = 1 - distribution
  x = np.round(1 / inv_distribution)
  x = np.where(distribution == 0, 0, x)
  return x.astype(int)


class DataIterator(Data):

  ''' Vertically merge several data object for iteration
  '''

  def __init__(self, data):
    super(DataIterator, self).__init__()
    if not isinstance(data, (tuple, list)):
      data = (data,)
    # ====== validate args ====== #
    if any(not isinstance(i, Data) for i in data):
      raise ValueError('data must be instance of MmapData or Hdf5Data, '
                       'but given data have types: {}'
                       ''.format(map(lambda x: str(type(x)).split("'")[1],
                                    data)))
    shape = data[0].shape[1:]
    if any(i.shape[1:] != shape for i in data):
      raise ValueError('all data must have the same trial dimension, but'
                       'given shape of all data as following: {}'
                       ''.format([i.shape for i in data]))
    # ====== defaults parameters ====== #
    self._data = data
    self._sequential = False
    self._distribution = [1.] * len(data)

  # ==================== properties ==================== #
  @property
  def shape(self):
    return (len(self),) + self._data[0].shape[1:]

  @property
  def array(self):
    start = self._start
    end = self._end
    idx = [(_apply_approx(i.shape[0], start), _apply_approx(i.shape[0], end))
           for i in self._data]
    idx = [(j[0], int(round(j[0] + i * (j[1] - j[0]))))
           for i, j in zip(self._distribution, idx)]
    return np.vstack([i[j[0]:j[1]] for i, j in zip(self._data, idx)])

  def __len__(self):
    start = self._start
    end = self._end
    return sum(round(i * (_apply_approx(j.shape[0], end) - _apply_approx(j.shape[0], start)))
               for i, j in zip(self._distribution, self._data))

  @property
  def distribution(self):
    return self._distribution

  def __str__(self):
    s = ['====== Iterator: ======']
    # ====== Find longest string ====== #
    longest_name = 0
    longest_shape = 0
    for d in self._data:
      name = d.name
      dtype = d.dtype
      shape = d.shape
      longest_name = max(len(name), longest_name)
      longest_shape = max(len(str(shape)), longest_shape)
    # ====== return print string ====== #
    format_str = ('Name:%-' + str(longest_name) + 's  '
                  'dtype:%-7s  '
                  'shape:%-' + str(longest_shape) + 's  ')
    for d in self._data:
      name = d.name
      dtype = d.dtype
      shape = d.shape
      s.append(format_str % (name, dtype, shape))
    # ====== batch configuration ====== #
    s.append('Batch: %d' % self._batch_size)
    s.append('Sequential: %r' % self._sequential)
    s.append('Distibution: %s' % str(self._distribution))
    s.append('Seed: %s' % str(self._seed))
    s.append('Range: [%.2f, %.2f]' % (self._start, self._end))
    return '\n'.join(s)

  def __repr__(self):
    return self.__str__()

  # ==================== batch configuration ==================== #
  def set_mode(self, distribution=None, sequential=None):
    '''
    Parameters
    ----------
    distribution : str, list or float
        'up', 'over': over-sampling all Data
        'down', 'under': under-sampling all Data
        list: percentage of each Data in the iterator will be iterated
        float: the same percentage for all Data
    sequential : bool
        if True, read each Data one-by-one, otherwise, mix all Data

    '''
    if sequential is not None:
      self._sequential = sequential
    if distribution is not None:
      # upsampling or downsampling
      if isinstance(distribution, str):
        distribution = distribution.lower()
        if 'up' in distribution or 'over' in distribution:
          n = max(i.shape[0] for i in self._data)
        elif 'down' in distribution or 'under' in distribution:
          n = min(i.shape[0] for i in self._data)
        else:
          raise ValueError("Only upsampling (keyword: up, over) "
                           "or undersampling (keyword: down, under) "
                           "are supported.")
        self._distribution = [n / i.shape[0] for i in self._data]
      # real values distribution
      elif isinstance(distribution, (tuple, list)):
        if len(distribution) != len(self._data):
          raise ValueError('length of given distribution must equal '
                           'to number of data in the iterator, but '
                           'len_data={} != len_distribution={}'
                           ''.format(len(self._data), len(self._distribution)))
        self._distribution = distribution
      # all the same value
      elif isinstance(distribution, float):
        self._distribution = [distribution] * len(self._data)
    return self

  # ==================== main logic of batch iterator ==================== #
  def __iter__(self):
    def create_iteration():
      seed = self._seed; self._seed = None
      if seed is not None:
        rng = np.random.RandomState(seed)
      else: # deterministic RandomState
        rng = struct()
        rng.randint = lambda x: None
        rng.permutation = lambda x: slice(None, None)
      # ====== easy access many private variables ====== #
      sequential = self._sequential
      start, end = self._start, self._end
      batch_size = self._batch_size
      distribution = np.asarray(self._distribution)
      # shuffle order of data (good for sequential mode)
      idx = rng.permutation(len(self._data))
      data = self._data[idx] if isinstance(idx, slice) else [self._data[i] for i in idx]
      distribution = distribution[idx]
      shape = [i.shape[0] for i in data]
      # ====== prepare distribution information ====== #
      # number of sample should be traversed
      n = np.asarray([i * (_apply_approx(j, end) - _apply_approx(j, start))
                      for i, j in zip(distribution, shape)])
      n = np.round(n).astype(int)
      # normalize the distribution (base on new sample n of each data)
      distribution = n / n.sum()
      distribution = _approximate_continuos_by_discrete(distribution)
      # somehow heuristic, rescale distribution to get more benifit from cache
      if distribution.sum() <= len(data):
        distribution = distribution * 3
      # distribution now the actual batch size of each data
      distribution = (batch_size * distribution).astype(int)
      assert distribution.sum() % batch_size == 0, 'wrong distribution size!'
      # predefined (start,end) pair of each batch (e.g (0,256), (256,512))
      idx = list(range(0, batch_size + distribution.sum(), batch_size))
      idx = list(zip(idx, idx[1:]))
      # Dummy return to initialize everything
      yield None
      #####################################
      # 1. optimized parallel code.
      if not sequential:
        # first iterators
        it = [iter(dat.set_batch(bs, seed=rng.randint(10e8),
                                 start=start, end=end,
                                 shuffle_level=self._shuffle_level))
              for bs, dat in zip(distribution, data)]
        # iterator
        while sum(n) > 0:
          batch = []
          for i, x in enumerate(it):
            if n[i] <= 0:
              continue
            try:
              x = next(x)[:n[i]]
              n[i] -= x.shape[0]
              batch.append(x)
            except StopIteration: # one iterator stopped
              it[i] = iter(data[i].set_batch(distribution[i],
                  seed=rng.randint(10e8), start=start, end=end,
                  shuffle_level=self._shuffle_level))
              x = next(it[i])[:n[i]]
              n[i] -= x.shape[0]
              batch.append(x)
          # got final batch
          batch = np.vstack(batch)
          # no idea why random permutation is much faster than shuffle
          if self._shuffle_level > 0:
            batch = batch[rng.permutation(batch.shape[0])]
          # return the iterations
          for i, j in idx[:int(ceil(batch.shape[0] / batch_size))]:
            yield batch[i:j]
      #####################################
      # 2. optimized sequential code.
      else:
        # first iterators
        batch_size = distribution.sum()
        it = [iter(dat.set_batch(batch_size, seed=rng.randint(10e8),
                                 start=start, end=end,
                                 shuffle_level=self._shuffle_level))
              for dat in data]
        current_data = 0
        # iterator
        while sum(n) > 0:
          if n[current_data] <= 0:
            current_data += 1
          try:
            x = next(it[current_data])[:n[current_data]]
            n[current_data] -= x.shape[0]
          except StopIteration: # one iterator stopped
            it[current_data] = iter(data[current_data].set_batch(batch_size, seed=rng.randint(10e8),
                                start=start, end=end,
                                shuffle_level=self._shuffle_level))
            x = next(it[current_data])[:n[current_data]]
            n[current_data] -= x.shape[0]
          # shuffle x
          if self._shuffle_level > 0:
            x = x[rng.permutation(x.shape[0])]
          for i, j in idx[:int(ceil(x.shape[0] / self._batch_size))]:
            yield x[i:j]
    # ====== create and return the iteration ====== #
    it = create_iteration()
    next(it)
    return it

  # ==================== Slicing methods ==================== #
  def __getitem__(self, y):
    start = self._start
    end = self._end
    idx = [(_apply_approx(i.shape[0], start), _apply_approx(i.shape[0], end))
           for i in self._data]
    idx = [(j[0], int(round(j[0] + i * (j[1] - j[0]))))
           for i, j in zip(self._distribution, idx)]
    size = np.cumsum([i[1] - i[0] for i in idx])
    if isinstance(y, int):
      idx = _get_closest_id(size, y)
      return self._data[idx][y]
    elif isinstance(y, slice):
      return self.array[y]
    else:
      raise ValueError('No support for indices type={}'.format(type(y)))


def _get_closest_id(size, y):
  idx = 0
  for i, j in enumerate(size):
    if y >= j:
      idx = i + 1
  return idx



from os import path
from setuptools import find_packages, setup

_ODIN_VERSION_ = '0.2.0'

def get_tensorflow_version():
  import subprocess
  try:
    task = subprocess.Popen(["nvcc", "--version"],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    out = task.stdout.read()
    if "release 9.0" in str(out, 'utf-8'):
      return "tensorflow-gpu==1.13.1"
  except FileNotFoundError as e:
    pass
  return "tensorflow==1.13.1"

here = path.abspath(path.dirname(__file__))

long_description = \
'''
An end-to-end framework support multi-modal data processing
and fast prototyping of machine learning algorithm in form
of organized networks.
'''

setup(
    name='odin-ai',
    version=_ODIN_VERSION_,
    description="Deep learning for research and production",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/imito/odin-ai',
    author='Trung Ngo Trong',
    author_email='trung@imito.ai',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Multimedia :: Sound/Audio :: Speech',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
    ],
    keywords='tensorflow pytorch machine learning neural networks deep learning bayesian',
    packages=find_packages(exclude=['examples', 'examples/*', 'docs', 'tests']),
    # scripts=['bin/speech-augmentation', 'bin/speech-test'],
    setup_requires=['pip>=19.0'],
    install_requires=['numpy>=1.0.0',
                      get_tensorflow_version(),
                      'tensorflow-probability==0.6.0',
                      'tensorflow-datasets',
                      'six>=1.9.0',
                      'scikit-learn>=0.18.0',
                      'matplotlib>=3.0.0',
                      'decorator',
                      'tqdm',
                      'dill',
                      'pyyaml',
                      'pycrypto'],
    extras_require={
        'visualize': ['pydot>=1.2.4',
                      'colorama',
                      'seaborn'],
        'tests': ['pytest',
                  'pandas',
                  'requests'],
        'audio': ['soundfile',
                  'resampy'],
        'docs': ['sphinx',
                 'sphinx_rtd_theme']
    },
    zip_safe=False)



from __future__ import division, absolute_import, print_function

import os
import re
import sys
import time
import shutil
import inspect
import numbers
import warnings
from itertools import chain
from functools import wraps
from collections import OrderedDict, defaultdict, Mapping
from contextlib import contextmanager
from abc import ABCMeta, abstractmethod
from six.moves import zip, range, cPickle
from six import add_metaclass, types, string_types

import numpy as np

from odin import backend as K
from odin.autoconfig import randint
from odin.utils import (as_tuple, as_list, uuid, cache_memory, is_number,
                        is_string, is_path, is_primitives, ctext,
                        flatten_list, get_all_files, is_pickleable,
                        FuncDesc, dummy_formatter, type_path,
                        get_module_from_path, wprint)
from odin.backend.role import (add_roles, has_roles, Parameter, Weight, Bias,
                               TrainableParameter, NNOpOutput)
from odin.nnet.base_desc import VariableDesc

import tensorflow as tf
from tensorflow.python.ops import init_ops

# ===========================================================================
# Other helpers
# ===========================================================================
def _get_vars_footprint(vars):
  return ';'.join([v.name for v in vars])

# ===========================================================================
# Global NNOp manager
# ===========================================================================
def get_all_nnops(scope=None, op_type=None):
  """ Return a dictionary of (name, nnops) for all created NNOp """
  allops = list(NNOp._ALL_NNOPS.values())
  # ====== matching the name scope ====== #
  if scope is not None:
    if not is_string(scope):
      scope = scope.name
    allops = [o for o in allops if o.name[:len(scope)] == scope]
  # ====== matching the NNOp type ====== #
  if op_type is not None:
    op_type = [i for i in as_tuple(op_type)
               if is_string(op_type) or issubclass(op_type, NNOp)]
    allops = [o for o in allops
              if any(o.__class__.__name__ == t if is_string(t)
                     else isinstance(o, t)
                     for t in op_type)]
  # ====== sorted by created time (earlier first) ====== #
  allops = sorted(allops, key=lambda x: x.timestamp,
                  reverse=False)
  return allops

def _assign_new_nnop(nnop):
  if not isinstance(nnop, NNOp):
    raise ValueError("The new assigned NNOp must be instance of odin.nnet.NNOp "
                     ", but the given object has type: %s" %
                     str(type(nnop)))
  name = nnop.name
  if not nnop.is_initialized:
    raise RuntimeError("Given NNOp with name: '%s' has not been initialized"
                       % name)
  if name in NNOp._ALL_NNOPS:
    raise RuntimeError("Cannot created NNOp with duplicated name, another NNOp "
                       "of type: %s, and name: '%s' has already existed" %
                       (type(NNOp._ALL_NNOPS[name]), name))
  NNOp._ALL_NNOPS[name] = nnop

# ===========================================================================
# Arguments scope
# ===========================================================================
__ARGS_SCOPE_STACK = [defaultdict(dict)]

def get_args_scope():
  return __ARGS_SCOPE_STACK[-1].copy()

@contextmanager
def args_scope(*ops_kwargs, **kwargs):
  """Stores the default arguments for the given set of applied_nnops.

  For usage, please see examples at top of the file.

  Parameters
  ----------
  ops_kwargs : series of list or tuple
    Contain the mapping from (`list_or_single_NNOp`, `**kwargs`)
     - The `list_or_single_NNOp` can be represented by string (name
       of a specific NNOp, or name the NNOp class), class (instance
       of NNOp), object (instance of any class inherited NNOp),
       list or tuple (which is the list of all above option).
     - `**kwargs` is list or dictionary containing the tuple of
       (keyword, value) that will define the defaults for each op in
       the listed ops
  kwargs : **kwargs
    default keywoard arguments for all given `NNOp` in `ops_kwargs`

  Return
  ------
  the current_scope, which is a dictionary of {op: {arg: value}}

  Raises
  ------
  TypeError: if list_ops is not a list or a tuple.
  ValueError: if any op in list_ops has not be decorated with @add_arg_scope.

  Note
  ----
  if the name scope is given, an incremental ID is generated for
  duplicated NNOp instead of UUID.
  """
  new_scope = defaultdict(dict)
  for ops, kw in ops_kwargs:
    if isinstance(kw, (tuple, list)):
      kw = dict(kw)
    if not isinstance(ops, (tuple, list)):
      ops = [ops]
    for o in ops:
      new_scope[o].update(kw)
  # ====== update Arguments Scope ====== #
  # copy prevous scopes
  args_scope = defaultdict(dict)
  for i, j in __ARGS_SCOPE_STACK[-1].items():
    args_scope[i] = j.copy()
  # update new scopes
  for op, kw in new_scope.items():
    args_scope[op].update(kw)
  # add the default kwargs
  for kw in args_scope.values():
    kw.update(kwargs)
  __ARGS_SCOPE_STACK.append(args_scope)
  # ====== return the scope ====== #
  yield None
  # ====== reset everything ====== #
  __ARGS_SCOPE_STACK.pop()

# ===========================================================================
# NNOp scope
# ===========================================================================
# each element is a list [scope_name, current_id]
_NNOP_SCOPE_STACK = []

def get_nnop_scope():
  # each element is a list [scope_name, current_id]
  if len(_NNOP_SCOPE_STACK) == 0:
    return ['', uuid()]
  return _NNOP_SCOPE_STACK[-1]

@contextmanager
def nnop_context(scope, reuse=None):
  """ A more generalized version of `tensorflow.variable_scope`,
  this function will also set the NNOp scope so any NNOp created within
  given scope will be affected

  Parameters
  ----------
  scope: string
      the name of current scope, new NNOp will be created as "scope/name"
  reuse: bool
      whether reuse variables for tensorflow.variable_scope

  Example
  -------
  >>> X = K.variable(x, name='x')
  >>> with N.nnop_scope(scope='s1'):
  ...     f1 = N.Dense(8)
  ...     f1(X)
  ...     with N.nnop_scope(scope='s2'):
  ...         f2 = N.Dense(25)
  ...         f2(X)
  ...         with N.nnop_scope(scope='s1', prefix='S'):
  ...             f3 = N.Dense(12)
  ...             f3(X)
  >>> with N.nnop_scope(scope='s1/s2', prefix='S'):
  ...     f4 = N.Dense(num_units=13)
  ...     f4(X)
  >>> # f1: s1/Dense_0
  >>> # f2: s1/s2/Dense_0
  >>> # f3: s1/s2/Dense_S0
  >>> # f4 == f3
  >>> print(N.get_all_nnops(scope='s1/s2')) # contains 2 NNOp
  >>> print(N.get_all_nnops(scope='s1')) # contains 3 NNOp
  """
  if not is_string(scope) or len(scope) == 0:
    raise ValueError("`scope` must be string type, length > 0.")
  # ====== prepare Name Scope ====== #object
  curr_scope, curr_opID = get_nnop_scope()
  # NO duplicate scope
  if scope not in curr_scope:
    curr_scope = scope if len(curr_scope) == 0 else \
        curr_scope + '/' + scope
  # name scope
  _NNOP_SCOPE_STACK.append([curr_scope, 0])
  # ====== return the scope ====== #
  var_scope = tf.get_variable_scope().name
  # NO repeating the scope in variable scope
  if any(s == scope for s in var_scope.split('/')): # this may cause error
    yield curr_scope
  else:
    with tf.variable_scope(scope, reuse=reuse):
      yield curr_scope
  # ====== reset everything ====== #
  _NNOP_SCOPE_STACK.pop()

# ===========================================================================
# Main Ops
# ===========================================================================
class _NNOp_Meta(ABCMeta):
  """ This meta-class ensure the NNOp argument scope is applied

  Note
  ----
  you can only modify the arguments and kwarguments using __call__
  from MetaClass, not __new__ of instance class.
  """
  def __new__(mcs, name, bases, class_dict):
    private = {'T', 'apply', '__call__', '__getstate__', '__setstate__',
               '__getnewargs__', 'get', 'get_variable_nnop', '__setattr__',
               'input_shape', 'placeholders', 'last_output', 'apply'}
    if name != 'NNOp':
      for attr in private:
        if attr in class_dict:
          raise RuntimeError("[Class:%s]The behavior of NNOp is "
              "restricted to ensure properly operations, the following "
              "methods or properties cannot be override: '%s'" %
              (ctext(name, 'red'), ctext(attr, 'yellow')))
    return super().__new__(mcs, name, bases, class_dict)

  def __call__(clazz, *args, **kwargs):
    assert issubclass(clazz, NNOp), \
    "NNOpMeta should only be used for NNOp subclass"
    # getting the default arguments to check user intentionally override
    # default argument.
    sign = inspect.signature(clazz.__init__)
    # ignore the self argument
    default_args = OrderedDict([(n, p.default)
                                for i, (n, p) in enumerate(sign.parameters.items())
                                if i > 0 and
                                p.kind not in (inspect.Parameter.VAR_POSITIONAL,
                                               inspect.Parameter.VAR_KEYWORD)])
    # ====== update the current argument scope ====== #
    # get current scope
    key_name = [clazz, str(clazz), clazz.__name__]
    current_scope = get_args_scope()
    for n in key_name:
      if n in current_scope:
        default_args.update(current_scope[n])
    # ====== update user specified args and kwargs ====== #
    # update the new arguments into default arguments
    new_kwargs = OrderedDict([
        (name, args[i]) if i < len(args) else (name, default)
        for i, (name, default) in enumerate(default_args.items())])
    new_kwargs.update(kwargs)
    # ====== create new instance and __init__ if necessary ====== #
    # This will call NNOp.__new__ to create an instance of NNOP,
    # if it found a duplicated NNOp pre-defined, it will return
    # the defined NNOp instead.
    op = clazz.__new__(clazz, *[], **new_kwargs)
    if not hasattr(op, '_name'):
      raise ValueError("NNOp must be given a name when initialized.")
    # check if op already initialized
    if op.name not in NNOp._ALL_NNOPS:
      clazz.__init__(op, *[], **new_kwargs)
    return op

@add_metaclass(_NNOp_Meta)
class NNOp(NNOpOutput):
  """ Basics of all Neural Network operators

  Properties
  ----------
  name: str
      identity of the operator, this name is the scope for its operator
      and should be unique.

  T: NNOp
      transpose operator of this one (NOTE: some ops does not support
      transpose and raise NotImplementedError)

  parameters: list of variables
      list of all parameters associated with this operator scope

  Abstract
  --------
  _apply(self, X, **kwargs): resulted variables
      apply take a list of variables and custom parameters to compute
      output variables
  _initialize(self, **kwargs):
      create parameters

  Override
  --------
  _transpose(self): NNOp
      return another NNOp which is transposed version of this ops

  Note
  ----
  All NNOp are pickle-able!
  You must use: protocol=cPickle.HIGHEST_PROTOCOL when dump NNOp.
  if NNOp is applied to a list of inputs, it will process each input seperated.
  """
  _ALL_NNOPS = {}

  @classmethod
  def search(clazz, name, path=None, prefix='model'):
    """ This method search for any objects decorated or instance
    of given NNOp
    from given `path` with all script have given `prefix`

    Parameters
    ----------
    name : string
        specific name of finding object

    path : {string, list of string}
        single folder or list of folder (or file) to search
        for the desire module

    prefix : string
        prefix for filtering .py file
    """
    # ====== check path ====== #
    if path is None:
      possible_path = ['.', './models', './model', './.models', './.model']
      script_path = os.path.dirname(sys.argv[0])
      path = [os.path.join(script_path, p) for p in possible_path]
      path = [p for p in path if os.path.exists(p) and os.path.isdir(p)]
    elif not isinstance(path, (tuple, list)):
      path = [path]
    if len(path) == 0:
      raise ValueError("Cannot find any available directory that contain the "
                       "model script.")
    # ====== search for model ====== #
    all_errors = {}
    for p in path:
      model_func, errors = get_module_from_path(name, path=p, prefix=prefix,
                                               return_error=True)
      all_errors.update(errors)
      model_func = [f for f in model_func
                    if isinstance(f, clazz)]
    if len(model_func) == 0:
      print(
          ctext("The following Exception happened during loading the modules:",
            'lightred'))
      for fpath, error in all_errors.items():
        print(" ", fpath, ":", ctext(error, 'red'))
      raise ValueError("Cannot find any model creator function with name='%s' "
                       "at paths='%s'" % (name, '; '.join(path)))
    return model_func[0]

  def __new__(clazz, *args, **kwargs):
    """ This __new__ ensures no NNOp with duplicated name and type is
    created, hence, if it found a duplicated one, it will return the
    duplicated """
    # ====== cPickle call __new__ ====== #
    if len(args) == 1 and len(kwargs) == 0 and \
    (is_string(args[0]) and '[__name__]' in args[0]):
      name = args[0].replace('[__name__]', '')
      # Found predefined NNOP
      if name in NNOp._ALL_NNOPS:
        instance = NNOp._ALL_NNOPS[name]
        if not isinstance(instance, clazz):
          raise RuntimeError("Found duplicated NNOp with type: '%s', "
              "which is different from pickled type: '%s'" %
              (type(instance), clazz))
        return instance
      # just create new instance
      return super(NNOp, clazz).__new__(clazz)
    # ====== For first time create instance ====== #
    # update Op name if it is None
    name = kwargs.get('name', None)
    op_scope = get_nnop_scope()
    # ====== special case Lambda op ====== #
    if clazz == Lambda:
      func = kwargs['func']
      assert inspect.isfunction(func) or inspect.ismethod(func),\
      "func for odin.nnop.base.Lambda must be a callable function or method"
      full_path = [i
                   for i in func.__qualname__.split('.')
                   if '<locals>' not in i]
      func_scope = full_path[:-1]
      func_name = full_path[-1]
      if name is None:
        name = func_name + '_' + str(op_scope[1])
      else:
        name = name
    # ====== general case ====== #
    else:
      # automatic generate name
      if name is None:
        name = clazz.__name__ + '_' + str(op_scope[1])
      # regulation for the NNOp name
      elif is_string(name):
        if '/' in name or ':' in name:
          raise ValueError("NNOp cannot contain '\\' or ':', given name is: %s" % name)
      # exception no support for given type
      else:
        raise ValueError("`name` for NNOp must be string, function, but given "
                         "`name` with type: %s" % name)
    # ====== add the scope ====== #
    # add scope to name
    if len(op_scope[0]) > 0:
      name = op_scope[0] + '/' + name
      op_scope[1] += 1
    # ====== check duplicated Op name ====== #
    if name in NNOp._ALL_NNOPS:
      old_clazz = NNOp._ALL_NNOPS[name].__class__
      if clazz != old_clazz:
        raise RuntimeError("Found predefined NNOp with type: %s, but "
            "the new NNOp has type: %s" % (old_clazz, clazz))
      return NNOp._ALL_NNOPS[name]
    # ====== allocate new Op ====== #
    created_time = time.time()
    new_op = super(NNOp, clazz).__new__(clazz)
    new_op._name = name
    new_op._timestamp = created_time
    # this store spontaneous args and kwargs feed to apply()
    new_op._current_args = ()
    new_op._current_kwargs = {}
    # all save-able attributes of NNOp store here
    new_op._save_states = {'_name': name,
                           '_timestamp': created_time}
    return new_op

  def __init__(self, **kwargs):
    # mapping: name -> VariableDesc, or Primitives
    self._kwargs_desc = OrderedDict()
    # mapping: ','.join(id(tensor)) -> output
    self._cache_outputs = {}
    self._last_input_footprint = ''
    self._transpose_ops = None
    self._is_initialized = False
    # mapping: variable_name -> (tensorflow_name, 'tensor' or 'variable')
    self._variable_info = OrderedDict()
    # special flags to detect if cPickle called with protocol >= 2
    self._new_args_called = False
    # this is special tricks, the unpickled ops stay useless
    # until its variables are restored, but if we restore the
    # variable right away, it create a session and prevent
    # any possibility of running tensorflow with multiprocessing
    # => store the _restore_vars_path for later, and restore
    # the variable when the NNOp is actually in used.
    self._set_restore_info(None, False)

  # ==================== variable restoring ==================== #
  def _set_restore_info(self, vars_path, delete_after):
    self._restore_vars_path = vars_path
    self._delete_vars_folder = bool(delete_after)
    return self

  def _restore_variables(self):
    """ This method can be called anywhere to make sure
    the variable related to this NNOp is restored after
    pickling.
    """
    if hasattr(self, '_restore_vars_path') and \
    self._restore_vars_path is not None:
      folder_path = os.path.dirname(self._restore_vars_path)
      if os.path.exists(folder_path):
        K.restore_variables(self._restore_vars_path)
        # delete cached folder if necessary
        if self._delete_vars_folder:
          shutil.rmtree(folder_path)
      else:
        wprint("NNOp: '%s' cannot restore variables from path: '%s'"
               (self.name, folder_path))
      # reset info
      self._set_restore_info(None, False)

  # ==================== pickling method ==================== #
  def __getstate__(self):
    if not self._new_args_called:
      raise RuntimeError(
          "You must use argument `protocol=cPickle.HIGHEST_PROTOCOL` "
          "when using `pickle` or `cPickle` to be able pickling NNOp.")
    self._new_args_called = False
    # add nnops here so all related NNOps are saved
    return self._save_states, self.nnops

  def __setstate__(self, states):
    # ====== default attribute ====== #
    self._current_args = ()
    self._current_kwargs = {}
    self._cache_outputs = {}
    # ====== save states ====== #
    self._save_states, nnops = states
    for key, val in self._save_states.items():
      setattr(self, key, val)
    # ====== check exist NNOp ====== #
    if self.name not in NNOp._ALL_NNOPS:
      _assign_new_nnop(self)
    elif NNOp._ALL_NNOPS[self.name] != self:
      raise RuntimeError("Mismatch NNOp, two NNOps with the same name "
                         "are initizlied:\n%s\nis different from:\n%s" %
                         (str(NNOp._ALL_NNOPS[self.name]), str(self)))

  def __getnewargs__(self):
    self._new_args_called = True
    return ('[__name__]' + self.name,)

  # ==================== properties ==================== #
  @property
  def save_states(self):
    """ Save state is dictionary of attribute name -> object
    those will be saved during pickling

    Note
    ----
    This property return a copy of the dictionary, any
    modification won't take effect on NNOp
    """
    return dict(self._save_states)

  def get(self, name, nnop=False):
    """"Simple shortcut for getting defined Variable, or NNOp
    within the scope of this `NNOp`

    Parameters
    ----------
    name : string
        the name, or part of the name of the `Variable` or `NNOp`
    nnop : bool
        if you want to get a `NNOp` with given name instead of
        `Variable`
    """
    # ====== get variable ====== #
    if not nnop:
      if isinstance(name, bytes):
        name = str(name, 'utf-8')
      elif not is_string(name):
        raise ValueError("`name` must be string.")
      if name not in self._variable_info:
        raise ValueError("Variable with name: '%s' hasn't been created." % name)
      return self.get_variable_nnop(name=name)
    # ====== nnop ====== #
    for op in self.nnops:
      if name == op.name:
        return op
    raise ValueError("Cannot find `NNOp` with name: '%s'" % name)

  @cache_memory
  def get_variable_nnop(self, name, shape=None, initializer=None, roles=[]):
    """ Initialize and return the Variable or NNOp with given description

    Parameters
    ----------
    name: str
        name for the variable
    shape: tuple, list
        expected shape for given variable
    initializer: variable, numpy.ndarray, function
        specification for initializing the weights, if a function
        is given, the arguments must contain `shape`
    roles: `odin.backend.role.Role`
        categories of this variable
    """
    self._restore_variables() # restore variable first
    if name in self.__dict__:
      raise RuntimeError("name='%s' has been defined in dictionary of "
                         "NNOp: '%s', type: %s" %
                         (name, self.name, self.__class__.__name__))
    if shape is not None:
      # convert to tuple if needed
      shape = as_tuple(shape)
      if any(d <= 0 or d is None for d in shape):
        raise ValueError((
            "Cannot create param with a non-positive shape dimension. "
            "Tried to create param with shape=%r, name=%r") %
            (shape, name))
    #####################################
    # 1. looking for Defined variable.
    if initializer is None and shape is None:
      if name not in self._variable_info:
        raise ValueError("Cannot find variable with name: %s for NNOps "
            "with name: %s" % (name, self.name))
      var_name, t = self._variable_info[name]
      # get variable
      if t == 'variable':
        var = K.get_all_variables(full_name=var_name)
        if len(var) == 0:
          raise RuntimeError("Cannot find variable with name: %s" % var_name)
        var = var[0]
      # get tensor
      elif t == 'tensor':
        name, footprint = var_name
        op = K.get_all_operations(footprint=footprint)
        if len(op) == 0:
          raise RuntimeError("Cannot find any Op with given footprint: %s" % footprint)
        var = op[0]._outputs[int(name.split(':')[-1])]
      # get nnops, use current args and kwargs for initialization
      elif t == 'nnop':
        var = var_name(*self._current_args, **self._current_kwargs)
      # only care about the first variable
      return add_roles(var, roles)
    #####################################
    # 2. initializing function.
    create_new_var = False
    if is_string(initializer):
      var = K.get_all_variables(name=initializer)
      if len(var) == 0:
        var = K.get_all_tensors(name=initializer)
      if len(var) == 0:
        raise ValueError("Cannot find any variable or tensor with name: "
            "'%s' for the initializer." % initializer)
      var = var[0]
    # is instance of NNOp
    elif isinstance(initializer, NNOp):
      var = initializer
    # is a callable
    elif hasattr(initializer, '__call__'):
      var = initializer(shape)
      if isinstance(var, np.ndarray) or\
      isinstance(initializer, init_ops.Initializer):
        create_new_var = True
    # is a scalar
    elif is_number(initializer):
      var = np.full(shape=shape, fill_value=initializer, dtype='float32')
      create_new_var = True
    # numpy array
    elif isinstance(initializer, np.ndarray):
      var = initializer
      create_new_var = True
    # actual tensor
    else:
      var = initializer
    #####################################
    # 3. Numpy ndarray.
    if create_new_var:
      var = K.variable(var, shape=shape, name=name)
      self._variable_info[name] = (var.name, 'variable')
    #####################################
    # 4. Shared variable, just check the shape.
    elif K.is_variable(var):
      _shape = var.shape.as_list()
      if shape is not None and tuple(shape) != tuple(_shape):
        raise Exception('Require variable with shape=%s, but was given different '
                        'shape=%s, name:%s.' %
                        (str(shape), str(_shape), str(name)))
      self._variable_info[name] = (var.name, 'variable')
    #####################################
    # 5. expression, we can only check number of dimension.
    elif K.is_tensor(var):
      # We cannot check the shape here, Tensor (even shared
      # variables) do not have a fixed compile-time shape. We can check the
      # dimensionality though.
      # Note that we cannot assign a name here. We could assign to the
      # `name` attribute of the variable, but the user may have already
      # named the variable and we don't want to override this.
      if shape is not None and var.shape.ndims != len(shape):
        raise Exception("parameter with name=%s has %d dimensions, should be "
                        "%d" % (name, var.shape.ndims, len(shape)))
      self._variable_info[name] = ((var.name, K.get_operation_footprint(var.op)),
                                   'tensor')
    elif isinstance(var, NNOp):
      self._variable_info[name] = (var, 'nnop')
    #####################################
    # 6. Exception.
    else:
      print(initializer)
      raise RuntimeError("cannot initialize parameters "
                         "(name:%s - shape:%s - roles: %s): "
                         "the `initializer` is not a numpy array, "
                         "a Tensor expression, a call-able, "
                         "or variable name as string (given type: %s)" %
                         (name, shape, roles, str(type(initializer))))
    # ====== assign annotations ====== #
    if K.is_tensor(var):
      return add_roles(var, roles)
    elif isinstance(var, NNOp):
      return var
    else:
      raise ValueError("Unsupport for variable type: %s" %
          type(var).__name__)

  @property
  def name(self):
    return self._name

  @property
  def timestamp(self):
    return self._timestamp

  @property
  def T(self):
    """ Return new ops which is transpose of this ops """
    if not self.is_initialized:
      raise RuntimeError("NNOp with name:'%s' has not been initialized, "
          "call the Op on any input to first initialize input information."
          % self.name)
    if self._transpose_ops is None:
      self._transpose_ops = self._transpose()
      if not isinstance(self._transpose_ops, NNOp):
        raise ValueError("The _transposed method must return NNOp."
                         "but the returned object has type=%s" %
                         str(type(self._transpose_ops)))
      # hard-fix the name of transpose NNOp
      self._transpose_ops._name = self.name + '_T'
      # this is very brain twisted
      self._transpose_ops._transpose_ops = self
    return self._transpose_ops

  @property
  def variables(self):
    """ Get all variables related to this Op, which include:

     - Initialized Variables
     - Variables belong to related NNOp within this Op.
     - Variables belong to related Tensor within this Op.
     - Variables within the scope of this NNOp.

    Note
    ----
    initialize all variables (basically, all Variables
    within this NNOp scope are initialized when you call
    `NNOp.variables`.
    """
    # ====== # restore variable first ====== #
    self._restore_variables()
    # ====== get all variables ====== #
    global_vars = {v.name: v for v in K.get_all_variables()}
    all_vars = []
    tensors = []
    for alias, (name, vtype) in self._variable_info.items():
      if vtype == 'variable':
        all_vars.append(global_vars[name])
      elif vtype == 'tensor':
        tensors.append(self.get_variable_nnop(alias))
      elif vtype == 'nnop':
        all_vars += name.variables
    # all variables from tensor
    all_vars += K.ComputationGraph(tensors).variables
    # all variables from NNOp
    for op in self.nnops:
      all_vars += op.variables
    # all variables within the scope
    all_vars += K.get_all_variables(scope=self.name)
    all_vars = tuple(sorted(set(all_vars), key=lambda x: x.name))
    # exception ignore variable with name IsTraining__
    all_vars = [v for v in all_vars if 'IsTraining__' not in v.name]
    return all_vars

  @property
  def nb_variables(self):
    n = 0
    for p in self.variables:
      n += np.prod(p.shape.as_list()).astype('int32')
    return n

  @property
  def parameters(self):
    """ return all TensorVariables which have the PARAMETER role"""
    return [i for i in self.variables if has_roles(i, Parameter)]

  @property
  def trainable_parameters(self):
    """ return all TensorVariables which have the TrainableParameter role"""
    return [i for i in self.variables if has_roles(i, TrainableParameter)]

  @property
  def nb_parameters(self):
    n = 0
    for p in self.parameters:
      n += np.prod(p.shape.as_list()).astype('int32')
    return n

  @property
  def is_initialized(self):
    return self._is_initialized

  @property
  def variable_info(self):
    return dict(self._variable_info)

  @property
  def nnops(self):
    """ Return all NNOp belong to
      * the initialization of this Op
      * or within the scope of this Op.
    """
    ops = []
    for name, (var, vtype) in self._variable_info.items():
      if vtype == 'nnop':
        ops.append(var)
    ops += get_all_nnops(scope=self.name)
    # make sure all the nested NNOp has the same _restore_vars_path
    # TODO: this is not good idea, if nested NNOp also restore the
    # variables, it will restore variables multiple times.
    # if hasattr(self, '_restore_vars_path') and \
    # self._restore_vars_path is not None:
    #     for o in ops:
    #         if not hasattr(o, '_restore_vars_path') or \
    #         o._restore_vars_path is None:
    #             o._restore_vars_path = self._restore_vars_path
    #             o._delete_vars_folder = self._delete_vars_folder
    # ====== remove duplicate NNOp ====== #
    final_ops = []
    for o in ops:
      if o is not self and o not in final_ops:
        final_ops.append(o)
    return final_ops

  @property
  def placeholders(self):
    """ Create list of placeholder to represent inputs of this NNOp
    """
    x = [i.placeholder for i in self._kwargs_desc.values()
         if isinstance(i, VariableDesc)]
    return x[0] if len(x) == 1 else x

  def set_placeholder(self, name, plh):
    return self._kwargs_desc[name].set_placeholder(plh)

  @property
  def last_output(self):
    if self._last_input_footprint not in self._cache_outputs:
      raise RuntimeError("This NNOp has not been called, and contains "
          "no information about outputs.")
    return self._cache_outputs[self._last_input_footprint]

  @property
  def input_shape(self):
    """NOTE: this input shape is only inferred from last inputs to
    this NNOp,
    since the input argument can has default argument, this input
    shape can change after everytime you call the NNOp"""
    x = [tuple(i.shape.as_list())
         for i in as_tuple(self.placeholders)]
    return x[0] if len(x) == 1 else x

  @property
  def input_shape_map(self):
    return {k: tuple(v.placeholder.shape.as_list())
            for k, v in self._kwargs_desc.items()
            if isinstance(v, VariableDesc)}

  @property
  def output_shape(self):
    """NOTE: this input shape is only inferred from last inputs to
    this NNOp,
    since the input argument can has default argument, this input
    shape can change after everytime you call the NNOp"""
    output = self.last_output
    extract_shape = lambda x: tuple(x.shape.as_list()) \
        if hasattr(x, 'get_shape') else \
        (x.shape if hasattr(x, 'shape') else ())
    if isinstance(output, (tuple, list)):
      return [extract_shape(o) for o in output]
    elif isinstance(output, Mapping):
      return OrderedDict([(name, extract_shape(o))
          for name, o in output.items()])
    return extract_shape(output)

  def __setattr__(self, name, value):
    # this record all assigned attribute to pickle them later
    # check hasattr to prevent recursive loop at the beginning before
    # __init__ is called
    if hasattr(self, '_save_states'):
      if name not in ('_save_states', '_cache_outputs',
                      '_current_args', '_current_kwargs',
                      '_last_input_footprint'):
        if is_primitives(value, inc_ndarray=True,
                         exception_types=[NNOp, FuncDesc]) or \
        (hasattr(value, '__call__') and is_pickleable(value)):
          self._save_states[name] = value
    return super(NNOp, self).__setattr__(name, value)

  # ==================== abstract method ==================== #
  def _initialize(self):
    """ This function is only called once, for the first time you
    apply this Ops
    """
    return None

  @abstractmethod
  def _apply(self, X, **kwargs):
    raise NotImplementedError

  def _transpose(self):
    raise NotImplementedError

  # ==================== interaction method ==================== #
  def _check_input_arg(self, x, name):
    """Validate input argument to `apply` function

    Parameters
    ----------
    x : {tensor, primitive}
      given symbol for the argument
    name : string
      name of the argument

    Return
    ------
    tuple of (VariableDesc, raw_data)
    VariableDesc: can be None if only given a primitive data types
    raw_data: can be None if only give VariableDesc
    """
    desc = None
    data = None
    # ====== if given tensor, use the provided tensor ====== #
    if K.is_tensor(x) or isinstance(x, VariableDesc):
      desc = x if isinstance(x, VariableDesc) else\
          VariableDesc(shape=x, name=x.name.split(':')[0])
      # keywords argument
      if name not in self._kwargs_desc:
        self._kwargs_desc[name] = desc
      curr_desc = self._kwargs_desc[name]
      # validating
      if isinstance(curr_desc, VariableDesc) and not curr_desc.is_equal(desc):
        raise ValueError("Found stored argument with description: '%s', given new "
                         "argument with description: '%s'" % (str(curr_desc), str(desc)))
      # overriding primitive
      else:
        self._kwargs_desc[name] = desc
    # ====== if given raw data, use saved tensor with new data ====== #
    elif isinstance(x, np.ndarray):
      # keywords
      if name not in self._kwargs_desc:
        self._kwargs_desc[name] = VariableDesc(
            shape=x.shape, dtype=x.dtype,
            name='VarArg%d' % int(name[1:]) if '.' == name[0] else name)
      desc = self._kwargs_desc[name]
      # validating
      if desc.shape[1:] != x.shape[1:] or \
      np.dtype(desc.dtype) != np.dtype(x.dtype):
        raise ValueError("NNOp has input description: '%s', given "
                         "ndarray: shape=%s dtype=%s" %
                         (str(desc), x.shape, x.dtype))
      # set data
      data = x
    # ====== primitive, keep it simple ====== #
    elif is_primitives(x, inc_ndarray=False):
      self._kwargs_desc[name] = x
      desc = x
    # ====== Uknown input, ERROR ====== #
    else:
      raise ValueError("The input argument for NNOp can be: "
          "`Tensor`, `odin.nnet.VariableDesc`, and primitive types"
          " (string, number, boolean, None, numpy.ndarray, numpy.generic)."
          " But the given type is: name='%s' : value=`%s`)" %
          (name, str(x)))
    return (desc, data)

  def apply(self, *args, **kwargs):
    # ====== restore variable first ====== #
    self._restore_variables()
    # ====== self.name can contain Model varable scope, hence,
    # remove the scope here  ====== #
    op_name = self.name.split('/')[-1]
    with nnop_context(scope=op_name, reuse=self.is_initialized):
      # ====== processing argument information ====== #
      # auto omit `self` argument
      sign = inspect.signature(self._apply)
      arg_name = []
      default_kwargs = {}
      inc_var_pos = False
      inc_var_key = False
      for n, p in sign.parameters.items():
        if p.kind == inspect.Parameter.VAR_POSITIONAL:
          inc_var_pos = True
        elif p.kind == inspect.Parameter.VAR_KEYWORD:
          inc_var_key = True
        else:
          arg_name.append(n)
          if p.default != inspect.Parameter.empty:
            default_kwargs[n] = p.default
      num_args = len(arg_name)
      # adding kwargs_new in Order
      kwargs_new = OrderedDict()
      # varargs arguments is named with '.' at the beginning
      pos_name = ['.%d' % i for i in range(len(args) - num_args)] if inc_var_pos else []
      # kwargs name
      key_name = [name for name in kwargs.keys() if name not in arg_name] if inc_var_key else []
      # get all positional arguments
      for idx, name in enumerate(arg_name + pos_name + key_name):
        if idx < len(args):
          x = args[idx]
        elif name in kwargs:
          x = kwargs[name]
        elif name in default_kwargs:
          x = default_kwargs[name]
        else: # not specified, use saved arguments
          # wprint('Cannot find argument at index: %d, name: %s' % (idx, name))
          continue
        # validate and record name to self._kwargs_desc
        kwargs_new[name] = self._check_input_arg(x=x, name=name)
      # kwargs now contains: arg_name -> (VariableDesc, ndarray(or None))
      kwargs = kwargs_new
      # add missing slot from _kwargs_desc
      for name, var in self._kwargs_desc.items():
        if name not in kwargs:
          kwargs[name] = (var, None)
      # ====== create footprint for unique arguments identification ====== #
      # footprint created by concat argument name and its
      # object python ID (primitive arugment using str)
      footprint = ''
      # positional argument
      for name, (desc, dat) in sorted(kwargs.items(),
                                   key=lambda x: x[0]):
        footprint += name + ':'
        if isinstance(desc, VariableDesc): # Tensor types
          desc = desc.placeholder
          footprint += type_path(desc) + '_' + str(id(desc))
        else: # primitive types
          footprint += type_path(desc) + '_' + str(desc)
        footprint += '|'
      # ====== convert all information to new op args and kwargs ====== #
      # store current arguments
      self._current_args = []
      included_names = []
      # positional args
      for name in arg_name:
        desc, dat = kwargs[name]
        self._current_args.append(desc if is_primitives(desc, inc_ndarray=False) else
                                  desc.placeholder)
        included_names.append(name)
      # varargs
      for name in sorted([i for i in kwargs.keys() if '.' == i[0]],
                         key=lambda x: int(x[1:])):
        desc, dat = kwargs[name]
        self._current_args.append(desc if is_primitives(desc, inc_ndarray=False) else
                                  desc.placeholder)
        included_names.append(name)
      # kwargs
      self._current_kwargs = {
          name: desc if is_primitives(desc, inc_ndarray=False) else desc.placeholder
          for name, (desc, dat) in kwargs.items()
          if name not in included_names}
      given_data = {desc.placeholder: dat
                    for name, (desc, dat) in kwargs.items()
                    if isinstance(desc, VariableDesc) and dat is not None}
      # ====== initialize first ====== #
      if not self._is_initialized:
        # call NNOp initialize
        self._initialize()
        self._is_initialized = True
        # only assign new NNOp if it is initialized
        _assign_new_nnop(self)
      # ====== calculate and return outputs ====== #
      # automatically restore all variable within this NNOp scope
      self._restore_variables()
      # Recall cached output
      if footprint in self._cache_outputs:
        y = self._cache_outputs[footprint]
      else: # First time generate output given footprint
        y = self._apply(*self._current_args, **self._current_kwargs)
        # all roles to all outputs
        y = add_roles(variables=y, roles=self.__class__)
        # record cahced return
        self._cache_outputs[footprint] = y
      # check if op_data given, then evaluate to get the results.
      if len(given_data) > 0:
        # only need to make sure all variables initialized
        # if we need to evaluate some expressions.
        K.initialize_all_variables()
        y = K.eval(y, feed_dict=given_data)
    # ====== reset the current information ====== #
    self._current_args = ()
    self._current_kwargs = {}
    self._last_input_footprint = footprint
    return y

  def __call__(self, *args, **kwargs):
    return self.apply(*args, **kwargs)

  def __str__(self):
    # ====== get all attrs ====== #
    all_attrs = dir(self)
    padding = '  '
    print_attrs = {}
    for name in all_attrs:
      if '_' != name[0] and (len(name) >= 2 and '__' != name[:2]) and\
      'name' != name and 'is_initialized' != name:
        try:
          attr = getattr(self, name)
        except Exception:
          continue
        # check print-able type
        if is_primitives(attr, inc_ndarray=True) or \
        (hasattr(attr, '__call__') and not inspect.ismethod(attr)):
          print_attrs[name] = attr
    print_attrs = sorted(print_attrs.items(), key=lambda x: x[0])
    # ====== format the output ====== #
    ops_format = '<%s, name: %s, init: %s>\n'
    ops_format = ops_format % (ctext(self.__class__.__name__, 'cyan'),
                               ctext(self.name, 'MAGENTA'),
                               self._is_initialized)
    for i, j in print_attrs:
      if isinstance(j, NNOp): # special format for NNOp
        s = '\n' + '\n'.join([2 * padding + line
            for line in str(j).split('\n')])
      else: # other attributes
        s = dummy_formatter(j)
      ops_format += padding + "%s: %s\n" % (ctext(i, 'yellow'), s)
    # ====== print tensor ====== #
    for name, (var, vtype) in self._variable_info.items():
      if vtype != 'tensor':
        continue
      v = self.get(name)
      roles = K.role.get_roles(v)
      ops_format += padding + "(Tensor)%s shape=%s, type=%s\n, role=%s\n" % \
          (ctext(v.name.split(':')[0], 'yellow'),
           ctext(tuple(v.shape.as_list()), 'yellow'),
           ctext(v.dtype.base_dtype.name, 'yellow'),
           ctext(';'.join(roles), 'yellow'))
    # ====== print Variable ====== #
    for var in self.variables:
      name = var.name.split(':')[0]
      vtype = type(var).__name__
      shape = tuple(var.shape.as_list())
      roles = K.role.get_roles(var)
      dtype = var.dtype.base_dtype.name
      ops_format += padding + "(%s)%s shape=%s, type=%s, role=%s\n" % \
          (vtype, ctext(name, 'yellow'),
              ctext(shape, 'yellow'), ctext(dtype, 'yellow'),
              ctext(';'.join(roles), 'yellow'))
    # ====== print NNOps ====== #
    for op in self.nnops:
      name = op.name
      otype = type(op).__name__
      ops_format += padding + '(NNOp)%s type=%s, inshape=%s\n' % \
          (ctext(name, 'yellow'),
              ctext(otype, 'yellow'),
              ctext(op.input_shape, 'yellow'))
    # ====== Input info ====== #
    for key, arg in self._kwargs_desc.items():
      if isinstance(arg, VariableDesc):
        arg = str(arg)
      else:
        arg = 'type:%s, value:%s' % (ctext(type_path(arg), 'cyan'),
                                     dummy_formatter(arg))
      name = ctext('[%s]' % key, 'yellow')
      ops_format += padding + name + arg + '\n'
    return ops_format[:-1]

  # ==================== Slicing ==================== #
  def __getitem__(self, key):
    return NNSliceOp(self, key)


_PRIMITIVE_TYPES = (tuple, list, dict, string_types, type(True),
                    types.FunctionType, numbers.Number, type(None),
                    init_ops.Initializer, NNOp, VariableDesc, type)

# ===========================================================================
# Helper
# ===========================================================================
def _prepend_scope_nnop_tree(scope, op, parent=None):
  """ Add scope to the left of all uninitialized NNOp and its children

  This must be called at initialization, not during applying the NNOp
  """
  _ = op.name.split('/')
  op_name = _[-1]
  op_scope = _[:-1]
  op_children = op.nnops
  # ====== merge the scope ====== #
  scope = scope.split('/')
  final_scope = []
  # re-organize the scope of the new NNOp
  # - No duplicate scope
  # - prioritize scope of parent NNOp appeared first
  for s in scope + op_scope:
    if s not in final_scope:
      final_scope += [s]
  # no empty scope
  op_scope = '/'.join([i for i in final_scope if len(i) > 0])
  # ====== modification is possible ====== #
  if not op.is_initialized:
    new_name = op_scope + '/' + op_name
    # check duplicated NNOp already defined and initialized
    if new_name in NNOp._ALL_NNOPS:
      new_op = NNOp._ALL_NNOPS[new_name]
      if type(new_op) != type(op):
        raise RuntimeError("NNOp of type %s with name '%s' already defined,"
          "but given new NNOp with type %s" % (type(new_op), new_name, op))
      op = new_op
      # if `parent` is provided, modify parent NNOp list as well
      if parent is not None:
        parent._variable_info[op_name] = (new_op, 'nnop')
    # otherwise, just modify the name of newly created NNOp
    else:
      op._name = new_name
  # ====== NNOp already initialized, no change ====== #
  else:
    pass
  # ====== modify the children NNOp as well ====== #
  for o in op_children:
    _prepend_scope_nnop_tree(op_scope, o, parent=op)
  return op

class Container(NNOp):
  """ Container is NNOp that contain other NNOp """

  def __init__(self, **kwargs):
    super(Container, self).__init__(**kwargs)
    self.debug = 0

  def set_debug(self, debug_mode):
    """
    Parameters
    ----------
    debug_mode : {0, 1, 2}
        0 - turn of logging
        1 - print minimal logging
        2 - print detail logging of each NNOp
    """
    self.debug = int(debug_mode)
    return self

  def set_nnops(self, ops):
    # remove None values
    if isinstance(ops, (tuple, list)):
      ops = [o for o in ops if o is not None]
      ops = flatten_list(ops, level=None)
    ops = list(as_tuple(ops, t=NNOp))

    # add new NNOp using it name and ignore the scope
    for o in ops:
      # modify the name and scope
      o = _prepend_scope_nnop_tree(scope=self.name, op=o)
      # store the new NNOp
      self.get_variable_nnop(name=o.name.split('/')[-1], initializer=o)
    # final assignment
    self._apply_ops = ops
    return self

  @contextmanager
  def _debug_mode(self, *args, **kwargs):
    args_desc = [tuple(x.shape.as_list()) if hasattr(x, 'get_shape') else str(x)
                 for x in self._current_args]
    kwargs_desc = {
        k: tuple(v.shape.as_list()) if hasattr(v, 'get_shape') else str(v)
        for k, v in self._current_kwargs.items()}
    # ====== print debug ====== #
    if self.debug > 0:
      print('**************** Start: %s ****************' %
          ctext(self.name, 'cyan'))
      print("First input:", ctext(str(args_desc) + ' ' + str(kwargs_desc), 'yellow'))
    # ====== running ====== #
    self._debug_ops = []
    yield
    # ====== print each op ====== #
    if len(self._debug_ops) > 0:
      type_format = '%-' + str(max(len(type(o).__name__) for o in self._debug_ops)) + 's'
      name_format = '%-' + str(max(len(o.name) for o in self._debug_ops)) + 's'
      for op in self._debug_ops:
        if self.debug == 1:
          print('[' + type_format % op.__class__.__name__ + ']',
                ctext(name_format % op.name, 'cyan'),
                "-> %s" % ctext(op.output_shape, 'yellow'))
        elif self.debug >= 2:
          print(str(op))
    # ====== ending and return ====== #
    if self.debug > 0:
      print('**************** End: %s ****************' %
            ctext(self.name, 'cyan'))

  def _print_op(self, op):
    # print after finish the op at each step
    self._debug_ops.append(op)

class Lambda(NNOp):

  """
  Parameters
  ----------
  func : callable
      main lambda function

  """

  def __init__(self, func, name=None):
    super(Lambda, self).__init__()
    # check main function
    self.set_function(func, is_transpose=False)
    self.set_function(func, is_transpose=True)

  def set_function(self, func, is_transpose=False):
    if not hasattr(func, '__call__'):
      raise ValueError("func must be call-able for Lambda.")
    if not isinstance(func, FuncDesc):
      func = FuncDesc(func)
    if is_transpose:
      self._funcT = func
    else:
      self._func = func
    return self

  def _initialize(self):
    pass

  def _apply(self, *args, **kwargs):
    # ====== update additional specialized variable for this NNOp ====== #
    for name in self._variable_info.keys():
      if name not in kwargs:
        kwargs[name] = self.get(name)
    return self._func(*args, **kwargs)

  def _transpose(self):
    return Lambda(name=self._funcT)

class NNSliceOp(NNOp):

  def __init__(self, ops, slice):
    if not isinstance(ops, NNOp):
      raise ValueError('ops must be instance of NNOp, but was given argument '
                       'has %s' % str(type(ops)))
    super(NNSliceOp, self).__init__()
    self._ops = ops
    if not isinstance(slice, (tuple, list)):
      slice = [slice]
    self.slice = slice

  def _apply(self, X, **kwargs):
    y = self._ops.apply(X, **kwargs)
    return_list = True if isinstance(y, (tuple, list)) else False
    # apply slice and calculate the shape
    output = [i[self.slice] for i in as_tuple(y)]
    # return output
    if return_list:
      return output
    return output[0]

  def __str__(self):
    ops_format = '<ops: %s, name: %s, init: %s, slice: %s>'
    return ops_format % (self._ops.__class__.__name__, self._ops.name,
                         self._ops.is_initialized, str(self.slice))

# ===========================================================================
# Simple ops
# ===========================================================================
class Dense(NNOp):

  def __init__(self, num_units,
               W_init=init_ops.glorot_uniform_initializer(seed=randint()),
               b_init=init_ops.constant_initializer(0),
               activation=K.linear,
               **kwargs):
    super(Dense, self).__init__(**kwargs)
    self.activation = (K.linear if activation is None else activation)
    self.W_init = W_init
    self.b_init = b_init
    self.num_units = num_units

  # ==================== abstract methods ==================== #
  def _transpose(self):
    # create the new dense
    return Dense(num_units=self.input_shape[-1],
                 W_init=Lambda(func=tf.transpose,
                               var_init={'a': self.get('W')}),
                 b_init=None if self.b_init is None else 0.,
                 activation=self.activation)

  def _initialize(self):
    input_shape = self.input_shape
    shape = (input_shape[-1], self.num_units)
    self.get_variable_nnop(name='W', shape=shape, initializer=self.W_init,
                      roles=Weight)
    if self.b_init is not None:
      self.get_variable_nnop(initializer=self.b_init,
          shape=(self.num_units,), name='b', roles=Bias)

  def _apply(self, X):
    # calculate projection
    activation = K.dot(X, self.get('W'))
    # add the bias
    if self.b_init is not None:
      activation = activation + self.get('b')
    # Nonlinearity might change the shape of activation
    return self.activation(activation)



from __future__ import print_function, division, absolute_import

import os
import shutil
import pickle
from collections import OrderedDict, Mapping
from six.moves import zip, range, cPickle

import numpy as np

from odin.utils.crypto import md5_checksum
from odin.utils import (get_file, Progbar, is_string,
                        ctext, as_tuple, eprint, wprint,
                        is_callable, flatten_list, UnitTimer)
from odin.fuel.data import (MmapData, Hdf5Data, open_hdf5, get_all_hdf_dataset,
                            MAX_OPEN_MMAP, Data, as_data)
from odin.fuel.utils import MmapDict, SQLiteDict, NoSQL
from odin.fuel.recipe_base import FeederRecipe, RecipeList


__all__ = [
    'Dataset',
    'copy_dataset2',
    'h5_to_ds'
]

# ===========================================================================
# Utilities
# ===========================================================================
def h5_to_ds(inpath, outpath):
  """ Convert a folder contains hdf5 files to a Dataset
  """
  pass

def copy_dataset2(origin, destination,
                  indices_filter=None, data_filter=None,
                  override=False):
  # ====== prepare input ====== #
  if is_string(origin):
    origin = Dataset(origin, read_only=True)
    own_ds = True
  elif isinstance(origin, Dataset):
    own_ds = False
  # ====== pass ====== #
  ds = origin.copy(destination,
                   indices_filter=indices_filter,
                   data_filter=data_filter,
                   override=override)
  # ====== end and return ====== #
  if own_ds:
    origin.close()
  return ds

# ===========================================================================
# Helper
# ===========================================================================
def _infer_separator(path):
  all_sep = ('\t', ' ', ';', ',')
  with open(path, 'r') as f:
    line = f.readline()
    line = line.strip()
    orig_length = len(line)
    for s in all_sep:
      if s not in line:
        continue
      if 1 < len(line.split(s)) < orig_length:
        return s
    raise RuntimeError("CSV file with the first line: `%s`, "
                       "cannot match separator in known list: `%s`"
                       % (line, str(all_sep)))

_audio_ext = ('.3gp', '.aa', '.aac', '.aax', '.act', '.aiff',
              '.amr', '.ape', '.au', '.awb', '.dct', '.dss',
              '.dvf', '.flac', '.gsm', '.ivs', '.m4a', '.m4b',
              '.m4p', '.mmf', '.mp3', '.mpc', '.msv', '.nsf',
              '.ogg,', '.opus', '.raw', '.sln', '.tta', '.vox',
              '.wav', '.wma', '.wv', '.webm', '.sph', '.pcm')

_image_ext = ('.tif', '.tiff', '.gif', '.jpeg', '.jpg', '.jif',
              '.jfif', '.jp2', '.jpx', '.j2k', '.j2c', '.fpx',
              '.pcd', '.png', '.pdf')

_ignore_files = ('.DS_Store',)

def _parse_data_descriptor(path, read_only):
  """ Return mapping: name -> (dtype, shape, Data, path) """
  if not os.path.isfile(path):
    return None
  file_ext = os.path.splitext(path)[-1].lower()
  file_name = os.path.basename(path)
  # ====== ignore ====== #
  if os.path.basename(path) in _ignore_files:
    return None
  # ====== audio file ====== #
  if file_ext in _audio_ext:
    return [(file_name, ('audio', 'unknown', None, path))]
  # ====== image file ====== #
  if file_ext in _image_ext:
    return [(file_name, ('image', 'unknown', None, path))]
  # ====== text file .txt ====== #
  if file_ext in ('.txt',):
    return [(file_name, ('txt', 'unknown', None, path))]
  # ====== check if is csv file ====== #
  if file_ext in ('.csv', '.tsv'):
    sep = _infer_separator(path)
    data = []
    # read by manually open file much faster than numpy.genfromtxt
    with open(path, 'r') as f:
      for line in f:
        line = line.strip()
        data.append(line.split(sep))
      data = np.array(data, dtype=str)
    return [('.'.join(file_name.split('.')[:-1]),
             ('csv', data.shape, data, path))]
  # ====== check if a file is Data ====== #
  try:
    dtype, shape = MmapData.read_header(path, read_only=True,
                                        return_file=False)
    # shape[1:], because first dimension can be resize afterward
    return [(file_name, (dtype, shape, None, path))]
  except Exception as e: # cannot read the header of MmapData, maybe Hdf5
    try:
      f = open_hdf5(path, read_only=read_only)
      ds = get_all_hdf_dataset(f)
      data = [Hdf5Data(dataset=i, hdf=f) for i in ds]
      return [(str(i.name), (str(i.dtype), i.shape, i, i.path)) for i in data]
    except Exception as e:
      pass
  # ====== try to load pickle file if possible ====== #
  try: # try with unpickling
    with open(path, 'rb') as f:
      data = cPickle.load(f)
      shape_info = 0
      if hasattr(data, 'shape'):
        shape_info = data.shape
      elif hasattr(data, '__len__'):
        shape_info = len(data)
      return [(file_name, (str(data.dtype) if hasattr(data, 'dtype') else
                           type(data).__name__,
                           shape_info, data, path))]
  except cPickle.UnpicklingError as e:
    try: # try again with numpy load
      with open(path, 'rb') as f:
        data = np.load(f)
        return [(file_name,
        (str(data.dtype) if hasattr(data, 'dtype') else type(data).__name__,
         len(data) if hasattr(data, '__len__') else 0, data, path))]
    except Exception as e:
      pass
  # ====== load memmap dict ====== #
  try:
    data = MmapDict(path, read_only=read_only)
    return [(file_name, ('memdict', len(data), data, path))]
  except Exception as e:
    pass
  # ====== load SQLiteDict ====== #
  if '.db' in os.path.splitext(path)[1]:
    try:
      db = SQLiteDict(path, read_only=read_only)
      name = os.path.basename(path).replace('.db', '')
      return [(tab if tab != SQLiteDict._DEFAULT_TABLE else name,
               ('sqlite', len(db.set_table(tab)), db.as_table(tab), path))
              for tab in db.get_all_tables()]
    except Exception as e:
      pass
  # ====== unknown datatype ====== #
  return [(file_name, ('unknown', 'unknown', None, path))]


# ===========================================================================
# Datasets
# ===========================================================================
class Dataset(object):
  """ This Dataset can automatically parse memmap (created by MmapData),
  MmapDict, pickled dictionary and hdf5 files and keep tracking the changes.

  Any file name with "readme" prefix will be parsed as text and showed as
  readme.

  Support data type:
   - .txt or .csv files:
   -

  Note
  ----
  for developer: _data_map contains: name -> (dtype, shape, Data or pathtoData)
  readme included with the dataset should contain license information
  All the file with `.db` extension will be treat as SQLite data
  """

  __INSTANCES = {}

  def __new__(clazz, *args, **kwargs):
    path = kwargs.get('path', None)
    if path is None:
      path = args[0]
    if not is_string(path):
      raise ValueError("`path` for Dataset must be string, but given "
                       "object with type: %s" % type(path))
    path = os.path.abspath(path)
    # Found old instance
    if path in Dataset.__INSTANCES:
      return Dataset.__INSTANCES[path]
    # new Dataset
    new_instance = super(Dataset, clazz).__new__(clazz)
    Dataset.__INSTANCES[path] = new_instance
    return new_instance

  def __init__(self, path, read_only=False, override=False):
    path = os.path.abspath(path)
    self.read_only = read_only
    self._readme_info = [ctext('README:', 'yellow'),
                         '------',
                         '  No information!']
    self._readme_path = None
    # flag to check cPickle called with protocol 2
    self._new_args_called = False
    # parse all data from path
    if path is not None:
      if override and os.path.exists(path) and os.path.isdir(path):
        shutil.rmtree(path)
        print('Overrided old dataset at path:', path)
      if os.path.isfile(path) and '.zip' in os.path.basename(path):
        self._load_archive(path,
            extract_path=path.replace(os.path.basename(path), ''))
      else:
        self._set_path(path, self.read_only)
    else:
      raise ValueError('Invalid path for Dataset: %s' % path)

  def _set_path(self, path, read_only):
    MAXIMUM_README_LINE = 25
    # all files are opened with default_mode=r+
    self._data_map = OrderedDict()
    self._path = os.path.abspath(path)
    self._default_hdf5 = os.path.basename(self._path) + '_default.h5'
    # svaed feeder info
    self._saved_indices = {}
    self._saved_recipes = {}
    # just make new dir
    if not os.path.exists(path):
      os.mkdir(path)
      os.mkdir(self.recipe_path)
      os.mkdir(self.index_path)
      return # not thing to do more
    elif not os.path.isdir(path):
      raise ValueError('Dataset path must be a folder.')
    # ====== Load all Data ====== #
    files = os.listdir(path)
    for fname in files:
      # found README
      if 'readme' == fname[:6].lower():
        readme_path = os.path.join(path, fname)
        with open(readme_path, 'r') as readme_file:
          readme = readme_file.readlines()[:MAXIMUM_README_LINE]
          readme = ['  ' + i[:-1] for i in readme if len(i) > 0 and i != '\n']
          readme.append(' => For more information: ' + readme_path)
          self._readme_info = [ctext('README:', 'yellow'),
                               '------'] + readme
          self._readme_path = readme_path
      # parse data
      data = _parse_data_descriptor(os.path.join(path, fname),
                                    read_only)
      if data is None: continue
      for key, d in data:
        if key in self._data_map:
          raise ValueError('Found duplicated data with follow info: '
                           '{}'.format(key))
        else:
          self._data_map[key] = d
    # ====== Load stored recipes ====== #
    # check recipes path
    if os.path.exists(self.recipe_path) and os.path.isfile(self.recipe_path):
      raise RuntimeError("Found a file at path: '%s', which supposed to "
                         "be a folder used for saving `recipe`"
                         % self.recipe_path)
    if not os.path.exists(self.recipe_path):
      os.mkdir(self.recipe_path)
    # all recipes is pickle-able
    for recipe_name in os.listdir(self.recipe_path):
      with open(os.path.join(self.recipe_path, recipe_name), 'rb') as f:
        recipe = cPickle.load(f)
        self._saved_recipes[recipe_name] = recipe
    # ====== load stored indices ====== #
    # check indices path
    if os.path.exists(self.recipe_path) and os.path.isfile(self.recipe_path):
      raise RuntimeError("Found a file at path: '%s', which supposed to "
                         "be a folder used for saving `index`"
                         % self.recipe_path)
    if not os.path.exists(self.index_path):
      os.mkdir(self.index_path)
    # load all saved indices
    for index_name in os.listdir(self.index_path):
      path = os.path.join(self.index_path, index_name)
      index = MmapDict(path=path, read_only=True)
      # remove extension from the name
      self._saved_indices[index_name] = index

  # ==================== Pickle ==================== #
  def __getstate__(self):
    if not self._new_args_called:
      raise RuntimeError(
          "You must use argument `protocol=cPickle.HIGHEST_PROTOCOL` "
          "when using `pickle` or `cPickle` to be able pickling Dataset.")
    self._new_args_called = False
    return self.path, self.read_only

  def __setstate__(self, states):
    path, read_only = states
    self._new_args_called = False
    self._set_path(path, read_only)

  def __getnewargs__(self):
    self._new_args_called = True
    return (self.path,)

  # ==================== archive loading ==================== #
  def _load_archive(self, path, extract_path):
    from zipfile import ZipFile, ZIP_DEFLATED
    try:
      zfile = ZipFile(path, mode='r', compression=ZIP_DEFLATED)
      allfile = zfile.namelist()
      # validate extract_path
      if not os.path.isdir(extract_path):
        raise ValueError('Extract path must be path folder, but path'
                         '={} is a file'.format(extract_path))
      extract_path = os.path.join(extract_path,
                                  os.path.basename(path).replace('.zip', ''))
      # found the extracted dir, use it
      if os.path.isdir(extract_path) and \
         set(os.listdir(extract_path)) == set(allfile):
        self._set_path(extract_path)
        return
      # decompress everything
      if not os.path.exists(extract_path):
        os.mkdir(extract_path)
      maxlen = max([len(i) for i in allfile])
      pb = Progbar(target=len(allfile), name="[Dataset] Loading Archive",
                   print_summary=True, print_report=True)
      for i, f in enumerate(allfile):
        zfile.extract(f, path=extract_path)
        pb['File'] = ('Unarchiving: %-' + str(maxlen) + 's') % f
        pb.add(1)
      # ====== finally set path ====== #
      self._set_path(extract_path)
    except IOError as e:
      raise IOError('Error loading archived dataset, path:{}, error:{}'
                    '.'.format(path, e))
    return None

  # ==================== properties ==================== #
  @property
  def basename(self):
    return os.path.basename(self.path)

  @property
  def path(self):
    return self._path

  @property
  def recipe_path(self):
    return os.path.join(self.path, 'recipe')

  @property
  def index_path(self):
    return os.path.join(self.path, 'index')

  @property
  def archive_path(self):
    """Return default archive path, which is:
        ../[dataset_name].zip
    """
    name = os.path.basename(self._path)
    return os.path.join(self._path, '..', name + '.zip')

  @property
  def md5(self):
    md5_text = ''
    for name, (dtype, shape, data, path) in sorted(self._data_map.items(),
                                                   key=lambda x: x[0]):
      md5_text += md5_checksum(path)
    return md5_text

  @property
  def size(self):
    """ return size in MegaByte"""
    size_bytes = 0
    for name, (dtype, shape, data, path) in self._data_map.items():
      try:
        size_bytes += os.path.getsize(path) # in bytes
      except Exception as e:
        eprint("Cannot acquire file size information, file: %s; error: %s"
               % (str(name), str(e)))
    return size_bytes / 1024. / 1024.

  def recipes(self):
    return {name: rcp
            for name, rcp in self._saved_recipes.items()}

  def indices(self):
    return {name: ids
            for name, ids in self._saved_indices.items()}

  def __len__(self):
    """ Return total number of data """
    return len(self._data_map)

  def __iter__(self):
    return self.items()

  def items(self):
    for name in self._data_map.keys():
      yield name, self.__getitem__(name)

  def iterinfo(self):
    """Return iteration of: (dtype, shape, loaded_data, path)"""
    for name, (dtype, shape, data, path) in self._data_map.items():
      yield (dtype, shape, self.__getitem__(name), path)

  def keys(self):
    """
    Return
    ------
    name of all Data
    """
    return self._data_map.keys()

  def values(self):
    """
    Return
    ------
    (dtype, shape, data, path) of Data
    """
    for k in self._data_map.keys():
      yield self.__getitem__(k)

  def archive(self):
    from zipfile import ZipFile, ZIP_DEFLATED
    path = self.archive_path
    zfile = ZipFile(path, mode='w', compression=ZIP_DEFLATED)

    files = set([_[-1] for _ in self._data_map.values()])

    prog = Progbar(target=len(files), name="[Dataset] Archiving",
                   print_report=True, print_summary=True)
    maxlen = max([len(os.path.basename(i)) for i in files])
    for i, f in enumerate(files):
      zfile.write(f, os.path.basename(f))
      prog['Data'] = ('Archiving: %-' + str(maxlen) + 's') \
          % os.path.basename(f)
      prog.add(1)
    zfile.close()
    return path

  # ==================== Feeder management ==================== #
  def add_indices(self, indices, name, override=False):
    # ====== validate name ====== #
    if not is_string(name):
      raise ValueError("`name` must be string, but given: %s" % str(type(name)))
    if name in self._saved_indices and not override:
      raise ValueError("Cannot override pre-defined INDEX with name: '%s'"
                      % name)
    # ====== validate indices ====== #
    path = os.path.join(self.index_path, name)
    ids = MmapDict(path)
    # predefined mapping, save or copy everything to a
    # MmapDict
    if isinstance(indices, Mapping):
      for name, (start, end) in indices.items():
        ids[name] = (start, end)
    # list of name, or (name, (start, end))
    elif isinstance(indices, (tuple, list, np.ndarray)):
      for i in indices:
        if is_string(i): # only name
          ids[i] = self['indices'][i]
        elif len(i) == 2: # name, (start, end)
          name, (start, end) = i
          ids[name] = (int(start), int(end))
        elif len(i) == 3: # name, start, end
          name, start, end = i
          ids[name] = (int(start), int(end))
        else:
          raise ValueError("Unsupport index parsing (name, start, end)"
                           "for: %s" % str(i))
    # flush everything to disk
    ids.flush(save_all=True)
    ids.close()
    # ====== assign new index ====== #
    self._saved_indices[name] = MmapDict(path, read_only=True)
    return self

  def remove_indices(self, name):
    if name is None: # remove all recipes
      for name in self._saved_indices.keys():
        self.remove_indices(name)
    # remove only a selected
    elif name in self._saved_indices:
      self._saved_indices[name].close()
      del self._saved_indices[name]
      os.remove(os.path.join(self.index_path, name))
    return self

  def add_recipes(self, recipes, name, override=False):
    """
    Parameters
    ----------
    """
    # ====== validate arguments ====== #
    if not is_string(name):
      raise ValueError("`name` must be string, but given: %s" % str(type(name)))
    if name in self._saved_recipes and not override:
      raise ValueError("Cannot override pre-defined RECIPE with name: '%s'"
                      % name)
    # ====== validate recipes list ====== #
    if isinstance(recipes, RecipeList):
      recipes = tuple(recipes._recipes)
    else:
      tmp = []
      for rcp in as_tuple(recipes, t=FeederRecipe):
        if isinstance(rcp, RecipeList):
          tmp += list(rcp._recipes)
        else:
          tmp.append(rcp)
      recipes = tuple(tmp)
    # ====== store the recipes to disk ====== #
    path = os.path.join(self.recipe_path, name)
    with open(path, 'wb') as f:
      cPickle.dump(recipes, f, protocol=cPickle.HIGHEST_PROTOCOL)
    # ====== update local recipes list ====== #
    self._saved_recipes[name] = recipes
    return self

  def remove_recipes(self, name):
    if name is None: # remove all recipes
      for name in self._saved_recipes.keys():
        self.remove_recipes(name)
    # remove only a selected
    elif name in self._saved_recipes:
      del self._saved_recipes[name]
      os.remove(os.path.join(self.recipe_path, name))
    return self

  def create_feeder(self, data, recipes, indices=None,
                    batch_filter=None, batch_mode='batch',
                    name=None, override=False):
    """
    Parameters
    ----------
    data: list of str
        list of name for all data used, the order of this
        list is the order of returned data.
    recipes: list or single odin.fuel.FeederRecipe
        the list of recipes defining the rule of transforming
        the data
    indices: None, string, dict, list
        list of (name, (start, end)) for iterating over files in Feeder
    batch_filter: call-able
        must be a function has take a list of np.ndarray as first arguments
        ([X]) or ([X, y]), you can return None to ignore given batch, return the
        data for accepting the batch
    batch_mode: 'batch' or 'file' (string type)
        'batch' mode return shuffling and return everything in small batches
        'file' mode return [(file_name, order_index, data...), ...]
    name: None, or string
        if name is provided, the feeder information will be saved,
        which include the `indices`, `recipes`

    Note
    ----
    by defaults, the Feeder is created using only 1 CPU with `buffer_size=1`
    using the method `set_multiprocessing(ncpu=None, buffer_size=None,
    maximum_queue_size=None)` for changing this information.
    """
    from odin.fuel.feeder import Feeder, IndexedData
    # check data
    data = [self.__getitem__(dat) if is_string(dat) else
            as_data(dat)
            for dat in as_tuple(data)]
    # check recipes
    if is_string(recipes):
      recipes = self._saved_recipes[recipes]
    else:
      recipes = as_tuple(recipes, t=FeederRecipe)
    # check indices
    if indices is None:
      indices = self.__getitem__('indices')
    elif is_string(indices):
      indices = self._saved_indices[indices]
    elif isinstance(indices, (Mapping, tuple, list, np.ndarray)):
      pass
    # ====== saving recipes and indices, if name is not None ====== #
    if is_string(name):
      if name not in self._saved_indices or override:
        self.add_indices(indices, name, override=True)
      if name not in self._saved_recipes or override:
        self.add_recipes(recipes, name, override=True)
    # ====== create Feeder ====== #
    feeder = Feeder(IndexedData(data=data, indices=indices),
                    batch_filter=batch_filter, batch_mode=batch_mode,
                    ncpu=1, buffer_size=1)
    return feeder.set_recipes(recipes)

  # ==================== Data management ==================== #
  def copy(self, destination,
           indices_filter=None, data_filter=None,
           override=False):
    """ Copy the dataset to a new folder and closed
    the old dataset

    """
    from distutils.dir_util import copy_tree
    read_only = self.read_only
    # indices
    if indices_filter is not None and \
    not is_callable(indices_filter) and \
    not isinstance(indices_filter, (tuple, list)):
      raise ValueError('`indices_filter` must be callable, tuple, list or None')
    if isinstance(indices_filter, (tuple, list)):
      tmp = tuple(indices_filter)
      indices_filter = lambda x: x in tmp
    # data name
    if data_filter is not None and \
    not is_callable(data_filter) and \
    not isinstance(data_filter, (tuple, list)):
      raise ValueError('`data_filter` must be callable, tuple, list or None')
    if isinstance(data_filter, (tuple, list)):
      tmp = tuple(data_filter)
      data_filter = lambda x: x in tmp
    # ====== other files which are not Data ====== #
    other_files = [i for i in os.listdir(self.path)
                   if i not in self]
    # ====== preprocessing ====== #
    destination = os.path.abspath(str(destination))
    if not os.path.exists(destination):
      os.mkdir(destination)
    elif not os.path.isdir(destination):
      raise ValueError('path at "%s" must be a folder' % destination)
    elif override:
      shutil.rmtree(destination)
      os.mkdir(destination)
    else:
      raise ValueError("A folder exist at path: '%s', cannot be overrided." %
                       destination)
    # ====== copy everything ====== #
    if indices_filter is None and data_filter is None:
      print("Copying %s files from '%s' to '%s' ..." %
        (ctext(len(self), 'cyan'),
         ctext(self.path, 'yellow'),
         ctext(destination, 'yellow')))
      copy_tree(self.path, destination)
    # ====== only data_filter ====== #
    elif indices_filter is None:
      data_list = [i for i in self.keys() if data_filter(i)]
      # copy all the data
      for name in data_list:
        org_path = os.path.join(self.path, name)
        dst_path = os.path.join(destination, name)
        print("Copying from '%s' to '%s' ..." %
              (ctext(org_path, 'yellow'),
               ctext(dst_path, 'yellow')))
        shutil.copy2(org_path, dst_path)
      # copy all the related indices
      for name in self.keys():
        org_path = os.path.join(self.path, name)
        dst_path = os.path.join(destination, name)
        if not os.path.exists(dst_path) and \
        ('indices' == name or any(i in data_list for i in name.split('_')[1:])):
          print("Copying Indices from '%s' to '%s'" % (ctext(org_path, 'cyan'),
                                                       ctext(dst_path, 'cyan')))
          shutil.copy2(org_path, dst_path)
    # ====== use indices_filter and data_filter ====== #
    else:
      if data_filter is None:
        all_data = list(self.keys())
      else:
        all_data = [i for i in self.keys()
                    if data_filter(i)]
      # list of data with separated indices
      separated_data = flatten_list(
          [k.split('_')[1:] for k in self.keys()
         if 'indices_' == k[:8]])
      # iterate over indices and copy one by one data
      for ids_name in [k for k in self.keys() if 'indices' == k[:7]]:
        indices = [(n, (s, e))
                   for n, (s, e) in self[ids_name]
                   if indices_filter(n)]
        # no match indices, skip
        if len(indices) == 0:
          continue
        nb_samples = sum(e - s for n, (s, e) in indices)
        # get all data assigned to given indices
        data = ids_name.split('_')[1:]
        if len(data) == 0:
          data = [i for i in all_data if i not in separated_data]
        else:
          data = [i for i in data if i in all_data]
        # if still no data found, skip
        if len(data) == 0:
          continue
        # copy each data
        for data_name in data:
          X = self[data_name]
          # copy big MmapDict
          if isinstance(X, MmapDict) and len(X) == len(self[ids_name]):
            new_path = os.path.join(destination, os.path.basename(X.path))
            print("Copying MmapDict from '%s' to '%s'" % (
                ctext(X.path, 'cyan'),
                ctext(new_path, 'cyan')))
            new_dict = MmapDict(new_path, cache_size=80000, read_only=False)
            for n, (s, e) in indices:
              new_dict[n] = X[n]
            new_dict.flush(save_all=True)
            new_dict.close()
          # copy MmapData
          elif isinstance(X, MmapData):
            Y = MmapData(path=os.path.join(destination, data_name),
                         dtype=X.dtype, shape=(0,) + X.shape[1:],
                         read_only=False)
            prog = Progbar(target=nb_samples,
                           print_report=True, print_summary=True,
                           name="Copying data: '%s' to path:'%s'" %
                           (ctext(data_name, 'yellow'),
                            ctext(Y.data_info, 'cyan')))
            for n, (s, e) in indices:
              Y.append(X[s:e])
              prog.add(e - s)
          # unknown data-type
          else:
            org_path = os.path.join(self.path, data_name)
            new_path = os.path.join(destination, data_name)
            # just copy directly the files
            if os.path.isfile(org_path) or \
            not os.path.exists(new_path):
              shutil.copy2(org_path, new_path)
              print("Copying '%s' to '%s' ..." %
                (ctext(org_path, 'cyan'), ctext(new_path, 'yellow')))
            else:
              wprint("Cannot copy: '%s' - %s" %
                (ctext(data_name, 'cyan'),
                 ctext(type(self[data_name]), 'yellow')))
        # copy the indices
        new_indices = MmapDict(os.path.join(destination, ids_name),
                               cache_size=80000, read_only=False)
        start = 0
        for n, (s, e) in indices:
          size = e - s
          new_indices[n] = (start, start + size)
          start += size
        new_indices.flush(save_all=True)
        new_indices.close()
    # ====== copy others files ====== #
    for f in other_files:
      org_path = os.path.join(self.path, f)
      dst_path = os.path.join(destination, f)
      if not os.path.exists(dst_path):
        if os.path.isdir(org_path): # directory
          copy_tree(org_path, dst_path)
        else: # single file
          shutil.copy2(org_path, dst_path)
    # ====== readme ====== #
    readme_name = os.path.basename(self._readme_path)
    dst_path = os.path.join(destination, readme_name)
    if not os.path.exists(dst_path):
      shutil.copy2(self._readme_path, dst_path)
    return Dataset(destination, read_only=read_only)

  def flush(self):
    for dtype, shape, data, path in self._data_map.values():
      if hasattr(data, 'flush'):
        data.flush()
      elif data is not None: # Flush pickling data
        with open(path, 'wb') as f:
          cPickle.dump(data, f, protocol=cPickle.HIGHEST_PROTOCOL)

  def close(self, name=None):
    # ====== close all Data ====== #
    if name is None: # close all files
      for name, (dtype, shape, data, path) in list(self._data_map.items()):
        if hasattr(data, 'close'):
          data.close()
        del data
        del self._data_map[name]
      # close all external indices and recipes
      for name, ids in self._saved_indices.items():
        ids.close()
      self._saved_indices.clear()
      for name, rcp in list(self._saved_recipes.items()):
        del rcp
      self._saved_recipes.clear()
      # Check if exist global instance
      if self.path in Dataset.__INSTANCES:
        del Dataset.__INSTANCES[self.path]
    # ====== close a particular file ====== #
    elif name in self._data_map:
      (dtype, shape, data, path) = self._data_map[name]
      if dtype == 'sqlite':
        data.sqlite.close()
      elif hasattr(data, 'close'):
        data.close()
      del data
      del self._data_map[name]

  # ==================== Some info ==================== #
  def _validate_memmap_max_open(self, name):
    # ====== check if MmapData excess limit, close 1 files ====== #
    if len(MmapData._INSTANCES) + 1 >= MAX_OPEN_MMAP:
      for i, (_dtype, _shape, _data, _path) in self._data_map.items():
        if isinstance(_data, MmapData) and i != name:
          self.close(name=i)
          self._data_map[i] = (_dtype, _shape, _path)
          break

  def __contains__(self, key):
    return key in self._data_map

  def find_prefix(self, feat_name, prefix):
    """ Specialized method for searching for Data or NoSQL
    with prefix, for example `prefix='indices'`:
      - `indices_%s_%s` % (feat1_name, feat2, ...)
    if no indices found, return the default indices with
    name 'indices'
    """
    indices = self[prefix] if prefix in self else None
    for key in self.keys():
      if prefix == key[:len(prefix)] and '_' + feat_name in key:
        indices = self[key]
    if indices is None:
      raise RuntimeError("Cannot find prefix: '%s' for feature with name: '%s', "
                         "all available name with given prefix are: %s" %
                         (prefix, feat_name, ','.join([k for k in self.keys()
                                                       if prefix == k[:len(k)]])
                         ))
    return indices

  def __getitem__(self, key):
    if is_string(key):
      if key not in self._data_map:
        raise KeyError('%s not found in this dataset' % key)
      dtype, shape, data, path = self._data_map[key]
      # return type is just a descriptor, create MmapData for it
      if data is None and \
      dtype is not 'unknown' and shape is not 'unknown':
        data = MmapData(path, read_only=self.read_only)
        self._data_map[key] = (data.dtype, data.shape, data, path)
        self._validate_memmap_max_open(key)
      return path if data is None else data
    raise ValueError('Only accept key type is string.')

  def get(self, key, default=None):
    if key in self._data_map:
      return self.__getitem__(key)
    return default

  def __setitem__(self, key, value):
    """
    Parameters
    ----------
    key : str or tuple
        if tuple is specified, it contain the key and the datatype
        which must be "memmap", "hdf5"
        for example: ds[('X', 'hdf5')] = numpy.ones((8, 12))
    """
    if not is_string(key) and not isinstance(key, (tuple, list)):
      raise ValueError('"key" is the name for Data and must be String or '
                       'tuple specified the name and datatype (memmap, hdf5).')
    # ====== check datatype ====== #
    datatype = 'memmap' # default datatype
    if isinstance(key, (tuple, list)):
      key, datatype = key
      datatype = str(datatype).lower()
      if datatype not in ('memmap', 'hdf5'):
        raise ValueError('datatype can only be "memmap" or "hdf5", but '
                         'the given data type is "%s"' % datatype)
    # ====== do nothing ====== #
    if key in self._data_map:
      return
    # ====== dict ====== #
    path = os.path.join(self.path, key)
    if isinstance(value, Mapping):
      if os.path.exists(path):
        raise Exception('File with path=%s already exist.' % path)
      d = MmapDict(path)
      for i, j in value.items():
        d[i] = j
      d.flush()
      # store new dict
      self._data_map[key] = (type(d).__name__, len(d), d, path)
    # ====== ndarray ====== #
    elif isinstance(value, np.ndarray):
      dtype, shape = value.dtype, value.shape
      if datatype == 'memmap':
        data = MmapData(path, dtype=dtype, shape=shape)
      else:
        path = os.path.join(self.path, self._default_hdf5)
        f = open_hdf5(path)
        data = Hdf5Data(key, hdf=f, dtype=dtype, shape=shape)
      # store new key
      self._data_map[key] = (data.dtype, data.shape, data, path)
      data[:shape[0]] = value
      # check maximum opened memmap
      self._validate_memmap_max_open(key)
    # ====== other types ====== #
    else:
      if os.path.exists(path):
        raise Exception('File with path=%s already exist.' % path)
      with open(path, 'wb') as f:
        cPickle.dump(value, f, protocol=cPickle.HIGHEST_PROTOCOL)
      # store new dict
      self._data_map[key] = (type(value).__name__,
                             len(value) if hasattr(value, '__len__') else 0,
                             value, path)

  def get_md5_checksum(self, excluded_name=[]):
    md5_text = ''
    all_data_items = {i: j
                      for i, j in self._data_map.items()
                      if i not in excluded_name}
    for name, (dtype, shape, data, path) in sorted(all_data_items.items(),
                                                   key=lambda x: x[0]):
      md5_text += md5_checksum(path)
    return md5_text

  def __str__(self):
    padding = '  '
    # NOTE: each element in the list is one line
    s = ['==========  ' +
         ctext('Dataset:%s Total:%d Size:%.2f(MB)', 'magenta') %
         (self.path, len(self._data_map), self.size) +
         '  ==========']
    s += self._readme_info
    s += [ctext('DATA:', 'yellow'),
          '----']
    # ====== Find longest string ====== #
    longest_name = 0
    longest_shape = 0
    longest_dtype = 0
    longest_file = 0
    print_info = []
    for name, (dtype, shape, data, path) in sorted(self._data_map.items()):
      shape = data.shape if hasattr(data, 'shape') else shape
      longest_name = max(len(name), longest_name)
      longest_dtype = max(len(str(dtype)), longest_dtype)
      longest_shape = max(len(str(shape)), longest_shape)
      longest_file = max(len(str(path)), longest_file)
      print_info.append([name, dtype, shape, path])
    # ====== return print string ====== #
    format_str = (padding + '%-' + str(longest_name + 2) + 's  '
                  '%-' + str(longest_dtype) + 's' + ctext(':', 'yellow') +
                  '%-' + str(longest_shape) + 's  '
                  'path:%-' + str(longest_file) + 's')
    for name, dtype, shape, path in print_info:
      s.append(format_str % ('"%s"' % name, dtype, shape, path))
    # ====== add recipes info ====== #
    for name, recipe in self._saved_recipes.items():
      s.append(ctext('(Recipe) ', 'yellow') + '"%s"' % name)
      for rcp in recipe:
        rcp = str(rcp)
        s.append('\n'.join([padding + line
                            for line in rcp.split('\n')]))
    # ====== add indices info ====== #
    for name, index in self._saved_indices.items():
      s.append(ctext('(Index) ', 'yellow') + '"%s"' % name)
      s.append(padding + str(index))
      name, (start, end) = next(index.items())
      s.append(padding + 'Sample: "%s %d-%d"' % (name, start, end))
    return '\n'.join(s)

  @property
  def readme(self):
    """ return text string of README of this dataset """
    if self._readme_path is not None:
      with open(self._readme_path, 'r') as f:
        readme = f.read()
    else:
      readme = self._readme_info[-1]
    return readme
